//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Data/IntensityDataItem.cpp
//! @brief     Implements class IntensityDataItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Data/IntensityDataItem.h"
#include "Base/Axis/IAxis.h"
#include "GUI/Model/Data/JobItemUtils.h"
#include "GUI/Model/Data/ProjectionItems.h"
#include "GUI/Model/Device/AxesItems.h"
#include "GUI/Model/Device/MaskItems.h"
#include "GUI/Model/Device/MaskUnitsConverter.h"
#include "GUI/Support/IO/ImportDataInfo.h"
#include "GUI/Support/XML/UtilXML.h"
#include "GUI/Util/Error.h"

namespace {

namespace Tag {

const QString BaseData("BaseData");
const QString ZAxis("ZAxis");
const QString Interpolation("Interpolation");
const QString Gradient("Gradient");

} // namespace Tag

// gradient map for colormaps
const QMap<QString, QCPColorGradient::GradientPreset> gradient_map = {
    {"Grayscale", QCPColorGradient::gpGrayscale},
    {"Hot", QCPColorGradient::gpHot},
    {"Cold", QCPColorGradient::gpCold},
    {"Night", QCPColorGradient::gpNight},
    {"Candy", QCPColorGradient::gpCandy},
    {"Geography", QCPColorGradient::gpGeography},
    {"Ion", QCPColorGradient::gpIon},
    {"Thermal", QCPColorGradient::gpThermal},
    {"Polar", QCPColorGradient::gpPolar},
    {"Spectrum", QCPColorGradient::gpSpectrum},
    {"Jet", QCPColorGradient::gpJet},
    {"Hues", QCPColorGradient::gpHues}};

QString startGradient = gradient_map.key(QCPColorGradient::gpThermal);

} // namespace

IntensityDataItem::IntensityDataItem()
    : DataItem(M_TYPE)
    , m_isInterpolated(true)
    , m_gradient(ComboProperty::fromList(gradient_map.keys(), startGradient))
    , m_zAxis(std::make_unique<AmplitudeAxisItem>())
{
    registerTag(T_MASKS, 0, -1, {MaskContainerItem::M_TYPE});
    setDefaultTag(T_MASKS);

    registerTag(T_PROJECTIONS, 0, -1, {ProjectionContainerItem::M_TYPE});
}

void IntensityDataItem::setDatafield(Datafield* data)
{
    if (data != nullptr) {
        if (data->rank() != 2)
            throw Error("Error in IntensityDataItem::setDatafield: cannot handle non-2D data");
        DataItem::setDatafield(data);
        updateAxesZoomLevel();
        updateAxesLabels();
        updateDataRange();
    } else
        DataItem::setDatafield(data);
}

double IntensityDataItem::xMin() const
{
    const double defaultXmin(0.0);
    return m_datafield ? m_datafield->axis(0).min() : defaultXmin;
}

double IntensityDataItem::xMax() const
{
    const double defaultXmax(1.0);
    return m_datafield ? m_datafield->axis(0).max() : defaultXmax;
}

double IntensityDataItem::yMin() const
{
    const double defaultYmin(0.0);
    return m_datafield ? m_datafield->axis(1).min() : defaultYmin;
}

double IntensityDataItem::yMax() const
{
    const double defaultYmax(1.0);
    return m_datafield ? m_datafield->axis(1).max() : defaultYmax;
}

double IntensityDataItem::lowerZ() const
{
    return zAxisItem()->min();
}

double IntensityDataItem::upperZ() const
{
    return zAxisItem()->max();
}

void IntensityDataItem::setLowerZ(double zmin)
{
    zAxisItem()->setMin(zmin);
    emit itemAxesRangeChanged();
}

void IntensityDataItem::setUpperZ(double zmax)
{
    zAxisItem()->setMax(zmax);
    emit itemAxesRangeChanged();
}

void IntensityDataItem::setLowerAndUpperZ(double zmin, double zmax)
{
    if (lowerZ() != zmin)
        setLowerZ(zmin);

    if (upperZ() != zmax)
        setUpperZ(zmax);
}

void IntensityDataItem::copyZRangeFromItem(DataItem* sourceItem)
{
    IntensityDataItem* source = dynamic_cast<IntensityDataItem*>(sourceItem);
    if (!source || source == this)
        return;
    setLowerZ(source->lowerZ());
    setUpperZ(source->upperZ());
}

QCPColorGradient IntensityDataItem::currentGradientQCP() const
{
    return gradient_map.value(currentGradient());
}

QString IntensityDataItem::currentGradient() const
{
    return gradientCombo().currentValue();
}

void IntensityDataItem::setCurrentGradient(const QString& gradient)
{
    m_gradient.setCurrentValue(gradient);
    emit gradientChanged();
}

ComboProperty IntensityDataItem::gradientCombo() const
{
    return m_gradient;
}

bool IntensityDataItem::isLog() const
{
    return zAxisItem()->isLogScale();
}

void IntensityDataItem::setLog(bool islog)
{
    zAxisItem()->setLogScale(islog);
}

bool IntensityDataItem::isInterpolated() const
{
    return m_isInterpolated;
}

void IntensityDataItem::setInterpolated(bool interp)
{
    m_isInterpolated = interp;
    emit interpolationChanged(interp);
}

bool IntensityDataItem::isZaxisLocked() const
{
    return m_zAxis->isLocked();
}

void IntensityDataItem::setZaxisLocked(bool state)
{
    return m_zAxis->setLocked(state);
}

void IntensityDataItem::updateCoords(const InstrumentItem* instrument)
{
    MaskUnitsConverter converter;
    converter.convertToNbins(this);

    GUI::Model::JobItemUtils::updateDataAxes(this, instrument);

    converter.convertFromNbins(this);
}

std::vector<int> IntensityDataItem::shape() const
{
    return {xSize(), ySize()};
}

void IntensityDataItem::reset(ImportDataInfo data)
{
    ASSERT(data.unitsLabel() == "nbins");
    ComboProperty combo = ComboProperty() << data.unitsLabel();
    setAxesUnitsCombo(combo);
    setXaxisTitle(data.axisLabel(0));
    setYaxisTitle(data.axisLabel(1));
    MaskUnitsConverter converter;
    converter.convertToNbins(this);
    setDatafield(std::move(data).intensityData().release());
    setAxesRangeToData();
    converter.convertFromNbins(this);
}

//! Sets zoom range of X,Y axes, if it was not yet defined.

void IntensityDataItem::updateAxesZoomLevel()
{
    // set zoom range of x-axis to min, max values if it was not set already
    if (upperX() < lowerX()) {
        setLowerX(xMin());
        setUpperX(xMax());
    }

    // set zoom range of y-axis to min, max values if it was not set already
    if (upperY() < lowerY()) {
        setLowerY(yMin());
        setUpperY(yMax());
    }

    const int nx = static_cast<int>(m_datafield->axis(0).size());
    xAxisItem()->setBinCount(nx);
    const int ny = static_cast<int>(m_datafield->axis(1).size());
    yAxisItem()->setBinCount(ny);
}

//! Init axes labels, if it was not done already.

void IntensityDataItem::updateAxesLabels()
{
    if (XaxisTitle().isEmpty())
        setXaxisTitle(QString::fromStdString(m_datafield->axis(0).axisName()));

    if (YaxisTitle().isEmpty())
        setYaxisTitle(QString::fromStdString(m_datafield->axis(1).axisName()));
}

//! Sets min,max values for z-axis, if axes is not locked, and ranges are not yet set.

void IntensityDataItem::updateDataRange()
{
    if (isZaxisLocked())
        return;

    computeDataRange();
    emit alignRanges();
}

void IntensityDataItem::computeDataRange()
{
    setLowerAndUpperZ(dataRange().first, dataRange().second);
}

//! Init zmin, zmax to match the intensity values range.
QPair<double, double> IntensityDataItem::dataRange() const
{
    const Datafield* data = datafield();

    const auto vec = data->flatVector();
    double min(*std::min_element(vec.cbegin(), vec.cend()));
    double max(*std::max_element(vec.cbegin(), vec.cend()));
    if (isLog()) {
        if (max > 10000) {
            min = 1.0;
            max = max * 1.1;
        } else {
            min = max / 10000;
            max = max * 1.1;
        }
    } else {
        max = max * 1.1;
    }

    return QPair<double, double>(min, max);
}

const AmplitudeAxisItem* IntensityDataItem::zAxisItem() const
{
    return m_zAxis.get();
}

AmplitudeAxisItem* IntensityDataItem::zAxisItem()
{
    return m_zAxis.get();
}

void IntensityDataItem::resetView()
{
    if (!m_datafield)
        return;

    setAxesRangeToData();
    if (!isZaxisLocked())
        computeDataRange();
}

void IntensityDataItem::writeTo(QXmlStreamWriter* w, const QString &projectDir) const
{
    XML::writeAttribute(w, XML::Attrib::version, uint(1));

    // parameters from base class
    w->writeStartElement(Tag::BaseData);
    DataItem::writeTo(w, projectDir);
    w->writeEndElement();

    // interpolation
    w->writeStartElement(Tag::Interpolation);
    XML::writeAttribute(w, XML::Attrib::value, m_isInterpolated);
    w->writeEndElement();

    // gradient
    w->writeStartElement(Tag::Gradient);
    m_gradient.writeTo(w);
    w->writeEndElement();

    // z axis
    w->writeStartElement(Tag::ZAxis);
    m_zAxis->writeTo(w);
    w->writeEndElement();
}

void IntensityDataItem::readFrom(QXmlStreamReader* r)
{
    const uint version = XML::readUIntAttribute(r, XML::Attrib::version);
    Q_UNUSED(version)

    while (r->readNextStartElement()) {
        QString tag = r->name().toString();

        // parameters from base class
        if (tag == Tag::BaseData) {
            DataItem::readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // interpolation
        } else if (tag == Tag::Interpolation) {
            XML::readAttribute(r, XML::Attrib::value, &m_isInterpolated);
            XML::gotoEndElementOfTag(r, tag);

            // gradient
        } else if (tag == Tag::Gradient) {
            m_gradient.readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // z axis
        } else if (tag == Tag::ZAxis) {
            m_zAxis->readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

        } else
            r->skipCurrentElement();
    }
}

void IntensityDataItem::writeNonSessionItems(QXmlStreamWriter* w, const QString& projectDir) const
{
    writeTo(w, projectDir);
}

void IntensityDataItem::readNonSessionItems(QXmlStreamReader* r)
{
    readFrom(r);
}

MaskContainerItem* IntensityDataItem::maskContainerItem()
{
    return dynamic_cast<MaskContainerItem*>(getItem(T_MASKS));
}

const MaskContainerItem* IntensityDataItem::maskContainerItem() const
{
    return dynamic_cast<MaskContainerItem*>(getItem(T_MASKS));
}

MaskContainerItem* IntensityDataItem::getOrCreateMaskContainerItem()
{
    MaskContainerItem* result = maskContainerItem();
    if (!result)
        result = model()->insertItem<MaskContainerItem>(this, -1, T_MASKS);
    return result;
}

ProjectionContainerItem* IntensityDataItem::projectionContainerItem()
{
    return dynamic_cast<ProjectionContainerItem*>(getItem(T_PROJECTIONS));
}

const ProjectionContainerItem* IntensityDataItem::projectionContainerItem() const
{
    return dynamic_cast<ProjectionContainerItem*>(getItem(T_PROJECTIONS));
}

ProjectionContainerItem* IntensityDataItem::getOrCreateProjectionContainerItem()
{
    ProjectionContainerItem* result = projectionContainerItem();
    if (!result)
        result = model()->insertItem<ProjectionContainerItem>(this, -1, T_PROJECTIONS);
    return result;
}

bool IntensityDataItem::hasMasks() const
{
    return (maskContainerItem() && !maskContainerItem()->maskItems().isEmpty());
}

bool IntensityDataItem::hasProjections() const
{
    return (projectionContainerItem() && projectionContainerItem()->hasChildren());
}
