//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Data/RealItem.cpp
//! @brief     Implements class RealItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Data/RealItem.h"
#include "Device/Data/DataUtils.h"
#include "GUI/Model/Data/IntensityDataItem.h"
#include "GUI/Model/Data/JobItemUtils.h"
#include "GUI/Model/Data/ProjectionItems.h"
#include "GUI/Model/Data/SpecularDataItem.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Device/MaskItems.h"
#include "GUI/Model/Model/SessionXML.h"
#include "GUI/Support/IO/AbstractDataLoader1D.h"
#include "GUI/Support/IO/DataLoaders1D.h"
#include "GUI/Support/IO/ImportDataInfo.h"
#include "GUI/Support/IO/ItemFileNameUtils.h"
#include "GUI/Support/XML/UtilXML.h"
#include "GUI/Util/DeserializationException.h"
#include "GUI/Util/Error.h"
#include "GUI/Util/MessageService.h"

namespace {
namespace Tag {

const QString InstrumentId("InstrumentId");
const QString Data("Data");
const QString Name("Name");
const QString NativeData("NativeData");
const QString BinaryData("BinaryData");
const QString NativeDataUnits("NativeDataUnits");

} // namespace Tag
} // namespace


RealItem::RealItem()
    : JobRealBase(M_TYPE)
{
    REFACTORED = true;
}

QString RealItem::realItemName() const
{
    return m_name;
}

void RealItem::setRealItemName(const QString& name)
{
    m_name = name;
    updateDataFileName();
}

DataItem* RealItem::dataItem() const
{
    return m_dataItem.get();
}

IntensityDataItem* RealItem::intensityDataItem() const
{
    return dynamic_cast<IntensityDataItem*>(dataItem());
}

SpecularDataItem* RealItem::specularDataItem() const
{
    return dynamic_cast<SpecularDataItem*>(dataItem());
}

void RealItem::initAsSpecularItem()
{
    const size_t rank = 1;
    initDataItem(rank, m_dataItem);
}

void RealItem::initAsIntensityItem()
{
    const size_t rank = 2;
    initDataItem(rank, m_dataItem);
}

bool RealItem::isIntensityData() const
{
    return intensityDataItem() != nullptr;
}

bool RealItem::isSpecularData() const
{
    return specularDataItem() != nullptr;
}

bool RealItem::hasNativeData() const
{
    return (nativeDataItem() != nullptr) && (nativeDataItem()->datafield() != nullptr);
}

DataItem* RealItem::nativeDataItem() const
{
    return m_nativeDataItem.get();
}

void RealItem::initNativeData()
{
    const size_t rank = isSpecularData() ? 1 : 2;
    initDataItem(rank, m_nativeDataItem);
    updateDataFileName();
}

void RealItem::removeNativeData()
{
    ASSERT(isSpecularData()); // not implemented for intensityDataItem

    if (nativeDataItem() != nullptr)
        nativeDataItem()->setDatafield(nullptr);
}

QString RealItem::nativeDataUnits() const
{
    return m_nativeDataUnits;
}

void RealItem::setNativeDataUnits(const QString& units)
{
    m_nativeDataUnits = units;
}

bool RealItem::holdsDimensionalData() const
{
    return nativeDataUnits() != "nbins";
}

const Datafield* RealItem::nativeDatafield() const
{
    return hasNativeData() ? nativeDataItem()->datafield() : nullptr;
}

//! takes ownership of data

void RealItem::setNativeDatafield(Datafield* data)
{
    nativeDataItem()->setDatafield(data); // takes ownership of odata
}

//! Sets Datafield to underlying item. Creates it if not existing.

void RealItem::setDatafield(Datafield* data)
{
    ASSERT(data && "Assertion failed in RealItem::setDatafield: passed data is nullptr");

    initDataItem(data->rank(), m_dataItem);

    dataItem()->setDatafield(data);
}

//! Sets imported data to underlying item. Creates it if not existing.
//! This is used for 1D import. For 2D import, use setDatafield.

void RealItem::setImportData(ImportDataInfo info)
{
    if (!info)
        return;

    const size_t data_rank = info.dataRank();
    initDataItem(data_rank, m_dataItem);
    initDataItem(data_rank, m_nativeDataItem);

    QString units_name = info.unitsLabel();
    auto output_data = info.clonedData();

    dataItem()->reset(std::move(info)); // TODO: Check this!
    setNativeDataUnits(units_name);
    m_nativeDataItem->setDatafield(output_data.release());

    updateDataFileName();
}

void RealItem::setNativeFileName(const QString& filename)
{
    m_nativeFileName = filename;
    updateDataFileName();
}

QString RealItem::nativeFileName() const
{
    return m_nativeFileName;
}

QString RealItem::instrumentId() const
{
    return m_instrumentId;
}

void RealItem::linkToInstrument(const InstrumentItem* instrument)
{
    if (instrument) {
        m_instrumentId = instrument->id();
        updateToInstrument(instrument);
    } else
        unlinkFromInstrument();
}

void RealItem::unlinkFromInstrument()
{
    m_instrumentId = QString{};
    updateToInstrument(nullptr);
}

bool RealItem::hasImportErrors() const
{
    return (dataLoader() != nullptr) ? dataLoader()->numErrors() > 0 : false;
}

void RealItem::setDataLoader(AbstractDataLoader* loader)
{
    m_dataLoader.reset(loader);
    if (m_dataLoader)
        connect(m_dataLoader.get(), &AbstractDataLoader::contentsProcessed, this,
                &RealItem::importContentsProcessed);
}

AbstractDataLoader* RealItem::dataLoader() const
{
    return m_dataLoader.get();
}

bool RealItem::rotationAffectsSetup() const
{
    if (!isIntensityData()) // rotation only for 2D items possible
        return false;

    const bool hasLinkToInstrument = !instrumentId().isEmpty();
    if (hasLinkToInstrument)
        return true;

    if (intensityDataItem()->hasMasks())
        return true;

    if (intensityDataItem()->hasProjections())
        return true;

    return false;
}

void RealItem::rotateData()
{
    if (!isIntensityData()) // rotation only for 2D items possible
        return;

    // -- first break instrument link, clear masks and projections
    unlinkFromInstrument();

    if (MaskContainerItem* maskContainer = intensityDataItem()->maskContainerItem())
        maskContainer->clear();

    if (ProjectionContainerItem* projectionsContainer =
            intensityDataItem()->projectionContainerItem())
        projectionsContainer->model()->removeRows(0, projectionsContainer->numberOfChildren(),
                                                  projectionsContainer->index());

    // -- now rotate data
    auto* const input = intensityDataItem()->datafield();
    intensityDataItem()->setDatafield(
        DataUtils::Data::createRearrangedDataSet(*input, 1).release());
    intensityDataItem()->setAxesRangeToData();
}

//! Updates the name of file to store intensity data.

void RealItem::updateDataFileName()
{
    if (DataItem* item = dataItem())
        item->setFileName(GUI::Model::FilenameUtils::realDataFileName(realItemName()));

    if (DataItem* item = nativeDataItem())
        item->setFileName(GUI::Model::FilenameUtils::nativeDataFileName(realItemName()));
}

// serialization/deserialization

QByteArray RealItem::serializeBinaryData() const
{
    QByteArray a;
    QDataStream s(&a, QIODevice::WriteOnly);
    s.setVersion(QDataStream::Qt_5_12);

    s << (quint8)1; // version
    s << m_nativeFileName;

    if (m_dataLoader) {
        s << m_dataLoader->persistentClassName();
        s << *m_dataLoader;
    } else
        s << QString(); // empty loader class name => no loader present

    return a;
}

void RealItem::deserializeBinaryData(const QByteArray& data)
{
    m_nativeFileName.clear();
    m_dataLoader.release();

    QDataStream s(data);
    s.setVersion(QDataStream::Qt_5_12);

    quint8 version;
    s >> version;

    if (version == 1) {
        s >> m_nativeFileName;
        QString persistentLoaderClassName;
        s >> persistentLoaderClassName;
        if (!persistentLoaderClassName.isEmpty()) {
            setDataLoader(
                DataLoaders1D::instance().createFromPersistentName(persistentLoaderClassName));
            // #baUserDefLoaders ++ implement userdefined loaders (may require new version!)
            if (!m_dataLoader)
                throw DeserializationException::tooNew();
            s >> *m_dataLoader;
            m_dataLoader->setRealDataItem(this);
        }
    } else
        throw DeserializationException::tooNew();

    if (s.status() != QDataStream::Ok)
        throw DeserializationException::streamError();
}

void RealItem::writeTo(QXmlStreamWriter* w, const QString& projectDir) const
{
    XML::writeAttribute(w, XML::Attrib::version, uint(1));

    // NOTE: The ordering of the XML elements is important in initialization

    // binary data
    QByteArray a = serializeBinaryData();
    if (!a.isEmpty()) {
        w->writeStartElement(Tag::BinaryData);
        XML::writeAttribute(w, XML::Attrib::binary_version, uint(1));
        w->writeCharacters(a.toBase64());
        w->writeEndElement();
    }

    // instrument id
    w->writeStartElement(Tag::InstrumentId);
    XML::writeAttribute(w, XML::Attrib::value, m_instrumentId);
    w->writeEndElement();

    // name
    w->writeStartElement(Tag::Name);
    XML::writeAttribute(w, XML::Attrib::value, m_name);
    w->writeEndElement();

    // native data units
    w->writeStartElement(Tag::NativeDataUnits);
    XML::writeAttribute(w, XML::Attrib::value, m_nativeDataUnits);
    w->writeEndElement();

    // data
    if (m_dataItem) {
        w->writeStartElement(Tag::Data);
        XML::writeItemAndChildItems(w, m_dataItem.get(), projectDir);
        w->writeEndElement();
    }

    // native data
    if (m_nativeDataItem) {
        w->writeStartElement(Tag::NativeData);
        XML::writeItemAndChildItems(w, m_nativeDataItem.get(), projectDir);
        w->writeEndElement();
    }
}

template <class T>
inline T* readItemToEnd(QXmlStreamReader* reader, SessionItem* parent_item, const QString& tag)
{
    return dynamic_cast<T*>(XML::readItemAndChildItems(reader, parent_item, tag));
}

QString RealItem::readFrom(QXmlStreamReader* r, const QString& projectDir,
                           MessageService* messageService)
{
    ASSERT(messageService);
    const uint version = XML::readUIntAttribute(r, XML::Attrib::version);
    Q_UNUSED(version)

    QString dataError, nativeDataError;
    while (r->readNextStartElement()) {
        QString tag = r->name().toString();

        // binary data
        if (tag == Tag::BinaryData) {
            const uint binary_version = XML::readUIntAttribute(r, XML::Attrib::binary_version);
            if (binary_version == 1) {
                QString valueAsBase64 = r->readElementText(QXmlStreamReader::SkipChildElements);
                const auto data = QByteArray::fromBase64(
                    valueAsBase64.toLatin1()); // #baimport add a unit test for this!
                deserializeBinaryData(data);
            }
            XML::gotoEndElementOfTag(r, tag);

            // instrument id
        } else if (tag == Tag::InstrumentId) {
            XML::readAttribute(r, XML::Attrib::value, &m_instrumentId);
            XML::gotoEndElementOfTag(r, Tag::InstrumentId);

            // name
        } else if (tag == Tag::Name) {
            XML::readAttribute(r, XML::Attrib::value, &m_name);
            XML::gotoEndElementOfTag(r, tag);

            // data
        } else if (tag == Tag::Data) {
            m_dataItem.reset(readItemToEnd<DataItem>(r, this, tag));
            dataError = m_dataItem->loadDatafield(messageService, projectDir);
            XML::gotoEndElementOfTag(r, tag);

            // native data
        } else if (tag == Tag::NativeData) {
            m_nativeDataItem.reset(readItemToEnd<DataItem>(r, this, tag));
            nativeDataError == m_nativeDataItem->loadDatafield(messageService, projectDir);
            XML::gotoEndElementOfTag(r, tag);

            // native data units
        } else if (tag == Tag::NativeDataUnits) {
            XML::readAttribute(r, XML::Attrib::value, &m_nativeDataUnits);
            XML::gotoEndElementOfTag(r, tag);

        } else
            r->skipCurrentElement();
    }

    // return error message
    if (dataError.isEmpty() && nativeDataError.isEmpty())
        return {};
    else if (!dataError.isEmpty() && nativeDataError.isEmpty())
        return (dataError);
    else if (dataError.isEmpty() && !nativeDataError.isEmpty())
        return (nativeDataError);
    else
        return (dataError + "\n" + nativeDataError);
}

void RealItem::copyTo(RealItem* const realdata_dst) const
{    
    // set the data members
    realdata_dst->m_nativeDataUnits = m_nativeDataUnits;
    realdata_dst->m_nativeFileName = m_nativeFileName;

    // update the instrument
    realdata_dst->unlinkFromInstrument();
    realdata_dst->m_instrumentId = m_instrumentId;

    // copy DataItems
    // NOTE: For consistency, the DataItem's parent is set to the destination RealItem
    // and its model is inherited from the destination RealItem.
    if (m_dataItem) {
        realdata_dst->m_dataItem.reset(
            dynamic_cast<DataItem*>(XML::copyItem(dataItem(), realdata_dst /* parent */)));
        realdata_dst->dataItem()->setParentAndModel(realdata_dst, realdata_dst->model());
        realdata_dst->dataItem()->setDatafield(dataItem()->datafield()->clone());
    }

    if (m_nativeDataItem) {
        realdata_dst->m_nativeDataItem.reset(
            dynamic_cast<DataItem*>(XML::copyItem(nativeDataItem(), realdata_dst /* parent */)));
        realdata_dst->nativeDataItem()->setParentAndModel(realdata_dst, realdata_dst->model());
        realdata_dst->nativeDataItem()->setDatafield(nativeDataItem()->datafield()->clone());
    }

    // discard DataLoader
    realdata_dst->m_dataLoader.release();
}

std::vector<int> RealItem::shape() const
{
    const auto* data_item = dataItem();
    if (!data_item) {
        ASSERT(data_item);
        return {};
    }
    return data_item->shape();
}

MaskContainerItem* RealItem::maskContainerItem()
{
    if (auto* intensity_data = intensityDataItem())
        return intensity_data->maskContainerItem();
    return nullptr;
}

//! Creates and inserts a data item except if such item with same tag already exists.
//! Checks for rank compatibility if already existing.
//! No further initialization (like clearing the data etc).

void RealItem::initDataItem(const std::size_t rank, std::unique_ptr<DataItem>& dataItem)
{
    // TODO: initDataItem can be replaced by a template which accepts only
    // SpecularDataItem or IntensityDataItem.

    ASSERT(rank == 1 || rank == 2);

    if (dataItem) {
        if ((rank == 1 && !dynamic_cast<const SpecularDataItem*>(dataItem.get()))
            || (rank == 2 && !dynamic_cast<const IntensityDataItem*>(dataItem.get())))
            throw Error("Error in RealItem::initDataItem: trying to set data "
                        "incompatible with underlying data item");
    } else {
        if (rank == 1)
            dataItem = std::make_unique<SpecularDataItem>();
        else // if rank == 2
            dataItem = std::make_unique<IntensityDataItem>();
        ASSERT(dataItem
               && "Assertion failed in RealItem::initDataItem: inserting data item failed.");
    }

    dataItem->setParentAndModel(this, model());
}

void RealItem::updateToInstrument(const InstrumentItem* instrument)
{
    DataItem* data_item = dataItem();
    if (!data_item)
        return;

    if (instrument) {
        GUI::Model::JobItemUtils::setDataItemCoords(data_item, instrument);
        if (dataItem())
            dataItem()->updateCoords(instrument);
        return;
    }

    // instrument == nullptr => unlinking => going back to native data
    if (isSpecularData()) {
        if (hasNativeData()) {
            std::unique_ptr<Datafield> native_data(nativeDataItem()->datafield()->clone());
            const QString units_label = nativeDataUnits();
            data_item->reset(ImportDataInfo(std::move(native_data), units_label));
        } else {
            specularDataItem()->setDatafield(nullptr);
        }
    } else {
        auto* native_data_item = nativeDataItem();
        auto* data_source = native_data_item ? native_data_item : data_item;

        std::unique_ptr<Datafield> native_data(data_source->datafield()->clone());
        const QString units_label = nativeDataUnits();
        data_item->reset(ImportDataInfo(std::move(native_data), units_label));
    }
}
