//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Model/ApplicationModels.cpp
//! @brief     Defines class holding all application models
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Model/ApplicationModels.h"
#include "GUI/Support/Data/SimulationOptionsItem.h"
#include "GUI/Support/XML/UtilXML.h"
#include "GUI/Util/DeserializationException.h"
#include <QXmlStreamWriter>

namespace {

namespace Tag {

const QString JobModel("JobModel");
const QString RealModel("RealModel");

} // namespace Tag

} // namespace

ApplicationModels::ApplicationModels()
{
    // creates and initializes models, order is important
    m_realDataModel = std::make_unique<RealModel>(this);
    m_jobModel = std::make_unique<JobModel>(this);

    connectModel(m_realDataModel.get());
    connectModel(m_jobModel.get());
}

RealModel* ApplicationModels::realModel() const
{
    return m_realDataModel.get();
}

JobModel* ApplicationModels::jobModel() const
{
    return m_jobModel.get();
}

void ApplicationModels::writeTo(QXmlStreamWriter* w, const QString& projectDir)
{
    // real model
    w->writeStartElement(Tag::RealModel);
    m_realDataModel->writeToProject(w, projectDir);
    w->writeEndElement();

    // job model
    w->writeStartElement(Tag::JobModel);
    m_jobModel->writeToProject(w, projectDir);
    w->writeEndElement();
}

void ApplicationModels::readFrom(QXmlStreamReader* r, const QString& projectDir,
                                 MessageService* messageService)
{
    try {
        QString tag = r->name().toString();

        // real model
        if (tag == Tag::RealModel) {
            m_realDataModel->readFromProject(r, projectDir, messageService);
            XML::gotoEndElementOfTag(r, tag);

            // job model
        } else if (tag == Tag::JobModel) {
            m_jobModel->readFromProject(r, projectDir, messageService);
            XML::gotoEndElementOfTag(r, tag);
        }
    } catch (DeserializationException& ex) {
        m_realDataModel->clear();
        m_jobModel->clear();
        r->raiseError(ex.text());
    }
}


QVector<DataItem*> ApplicationModels::dataItems() const
{
    ASSERT(m_realDataModel && m_jobModel);
    return QVector<DataItem*>() << m_realDataModel->dataItems() << m_jobModel->dataItems();
}

void ApplicationModels::connectModel(SessionModel* model) const
{
    if (!model)
        return;
    connect(model, &SessionModel::dataChanged, this, &ApplicationModels::modelChanged,
            Qt::UniqueConnection);
    connect(model, &SessionModel::rowsRemoved, this, &ApplicationModels::modelChanged,
            Qt::UniqueConnection);
    connect(model, &SessionModel::rowsInserted, this, &ApplicationModels::modelChanged,
            Qt::UniqueConnection);
}
