//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Model/RealModel.cpp
//! @brief     Implements class RealModel
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Model/RealModel.h"
#include "GUI/Model/Data/DataItem.h"
#include "GUI/Model/Data/RealItem.h"
#include "GUI/Support/XML/UtilXML.h"
#include "GUI/Util/Error.h"

namespace {
namespace Tag {

const QString RealData("RealData");

}
} // namespace

RealModel::RealModel(QObject* parent)
    : SessionModel("no_model_tag", parent)
{
    setObjectName("RealModel");
}

RealModel::~RealModel() = default;

QVector<DataItem*> RealModel::dataItems() const
{
    QVector<DataItem*> result;
    for (auto* realData : realItems()) {
        if (auto* intensityItem = realData->dataItem())
            result.push_back(intensityItem);
        if (auto* native_data_item = realData->nativeDataItem())
            result.push_back(native_data_item);
    }
    return result;
}

void RealModel::clear()
{
    for (auto* realItem : realItems())
        removeRealItem(realItem);
}

void RealModel::writeToProject(QXmlStreamWriter* w, const QString& projectDir)
{
    XML::writeAttribute(w, XML::Attrib::version, uint(1));

    // real items
    for (auto* realItem : realItems()) {
        w->writeStartElement(Tag::RealData);
        XML::writeAttribute(w, XML::Attrib::name, realItem->realItemName());
        realItem->writeTo(w, projectDir);
        w->writeEndElement();
    }
    deleteOldDataFiles(projectDir);
}

void RealModel::readFromProject(QXmlStreamReader* r, const QString& projectDir,
                                MessageService* messageService)
{
    clear();

    const uint version = XML::readUIntAttribute(r, XML::Attrib::version);
    Q_UNUSED(version)

    while (r->readNextStartElement()) {
        QString tag = r->name().toString();

        // real item
        if (tag == Tag::RealData) {
            createRealItem()->readFrom(r, projectDir, messageService);
            XML::gotoEndElementOfTag(r, tag);

        } else
            r->skipCurrentElement();
    }
    if (r->hasError())
        throw Error(r->errorString());
}

RealItem* RealModel::createRealItem()
{
    auto realItem = std::make_unique<RealItem>();
    realItem->setParentAndModel(nullptr, this);
    m_realItems.push_back(std::move(realItem));
    emit realDataAddedOrRemoved();
    return m_realItems.back().get();
}

RealItem* RealModel::insertSpecularDataItem()
{
    RealItem* r = createRealItem();
    r->initAsSpecularItem();
    return r;
}

RealItem* RealModel::insertIntensityDataItem()
{
    RealItem* r = createRealItem();
    r->initAsIntensityItem();
    return r;
}

RealItem* RealModel::insertDataItem(int rank)
{
    return (rank == 1) ? insertSpecularDataItem() : insertIntensityDataItem();
}

QVector<RealItem*> RealModel::realItems() const
{
    QVector<RealItem*> items(m_realItems.size());
    for (int i = 0; i < items.size(); i++)
        items[i] = m_realItems[i].get();
    return items;
}


QVector<RealItem*> RealModel::realItems(int rank) const
{
    return rank == 1 ? realItems1D() : realItems2D();
}

QVector<RealItem*> RealModel::realItems1D() const
{
    QVector<RealItem*> vec_1d;

    for (auto* item : realItems())
        if (item->isSpecularData())
            vec_1d.push_back(item);

    return vec_1d;
}

QVector<RealItem*> RealModel::realItems2D() const
{
    QVector<RealItem*> vec_2d;

    for (auto* item : realItems())
        if (item->isIntensityData())
            vec_2d.push_back(item);

    return vec_2d;
}

void RealModel::removeRealItem(RealItem* realItem)
{
    ASSERT(realItem);
    m_realItemFilesToDelete.append(realItem->dataItem()->fileName());
    int index = realItems().indexOf(realItem);
    m_realItems.erase(m_realItems.begin() + index);
    emit realDataAddedOrRemoved();
}

QStringList RealModel::realItemNames() const
{
    QStringList result;
    for (auto* item : realItems())
        result.append(item->realItemName());
    return result;
}

void RealModel::deleteOldDataFiles(const QString& projectDir)
{
    for(QString fileName : m_realItemFilesToDelete)
            QFile::remove(projectDir + "/" + fileName);
    m_realItemFilesToDelete.clear();
}
