//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Model/SessionXML.cpp
//! @brief     Implements reader and writer classes for SessionModel
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Model/SessionXML.h"
#include "GUI/Model/MakeItem/ItemFactory.h"
#include "GUI/Model/Model/SessionModel.h"
#include "GUI/Model/Sample/MaterialItem.h"
#include "GUI/Support/Data/SessionFlags.h"
#include "GUI/Support/Data/SessionItemTags.h"
#include "GUI/Support/XML/UtilXML.h"
#include "GUI/Util/ComboProperty.h"
#include "GUI/Util/DeserializationException.h"
#include "GUI/Util/Error.h"
#include "GUI/Util/MessageService.h"
#include <QColor>
#include <QUuid>
#include <QXmlStreamWriter>

namespace {

void report_error(MessageService* messageService, SessionModel* model, const QString& message)
{
    if (!messageService)
        throw Error(QString("Warning: ") + message);
    messageService->addWarning(model, message);
}

SessionItem* createItem(SessionItem* parent, const QString& modelType, const QString& tag)
{
    ASSERT(parent);
    if (parent->sessionItemTags()->isSingleItemTag(tag))
        return parent->getItem(tag);
    try {
        return parent->model()->insertNewItem(modelType, parent, -1, tag);
    } catch (const std::exception&) {
        return nullptr; // error will be reported later
    }
}

} // namespace


void XML::writeModel(QXmlStreamWriter* writer, SessionItem* modelRootItem)
{
    ASSERT(writer);
    ASSERT(modelRootItem);
    writer->writeStartElement(modelRootItem->model()->getModelTag());

    writeItemAndChildItems(writer, modelRootItem);

    writer->writeEndElement(); // m_model_tag
}

void XML::writeItemAndChildItems(QXmlStreamWriter* writer, const SessionItem* item, const QString& projectDir)
{
    ASSERT(item);
    if (item->itemParent()) {
        writer->writeStartElement(ItemTag);
        writer->writeAttribute(ModelTypeAttribute, item->modelType());
        QString tag = item->itemParent()->tagFromItem(item);
        writer->writeAttribute(Tags::TagAttribute, tag);
        writer->writeAttribute(Tags::DisplayNameAttribute,
                               item->roleProperty(SessionFlags::DisplayNameRole).toString());
        for (int role : item->getRoles()) {
            if (role == Qt::DisplayRole || role == Qt::EditRole)
                writeVariant(writer, item->value(), role);
        }
    }

    for (auto* child : item->itemChildren())
        if (item->allowWritingChildToXml(child))
            writeItemAndChildItems(writer, child, projectDir);

    QByteArray a = item->serializeBinaryData();
    if (!a.isEmpty()) {
        writer->writeStartElement(Tags::BinaryData);
        writer->writeAttribute(XML::Version, "1");
        writer->writeCharacters(a.toBase64());
        writer->writeEndElement();
    }

    QString nonSessionItemContent;
    QXmlStreamWriter nonSessionItemStream(&nonSessionItemContent);
    nonSessionItemStream.writeStartElement(Tags::NonSessionItemData); // for following writeAttrib
    const auto sizeBefore = nonSessionItemContent.size();
    item->writeNonSessionItems(&nonSessionItemStream, projectDir);
    if (sizeBefore != nonSessionItemContent.size()) {
        writer->writeStartElement(Tags::NonSessionItemData);
        item->writeNonSessionItems(writer, projectDir);
        writer->writeEndElement();
    }

    if (item->itemParent())
        writer->writeEndElement(); // ItemTag
}

void XML::readItems(QXmlStreamReader* reader, SessionItem* parent, QString topTag,
                    MessageService* messageService, const QString endTag)
{
    bool legacyDistributionFound = false;
    ASSERT(reader);
    ASSERT(parent);
    const QString start_type = parent->model()->getModelTag();

    while (!reader->atEnd()) {
        reader->readNext();
        if (reader->isStartElement()) {
            if (reader->name().toString() == XML::ItemTag) {
                const QString model_type =
                    reader->attributes().value(Tags::ModelTypeAttribute).toString();

                if (model_type == "ParticleDistribution") {
                    if (!legacyDistributionFound) {
                        QString message(
                            "This file contains particle distributions in the sample. Elements of "
                            "this type are not supported any more. Reading the file will be "
                            "continued as much as possible, but you should review the sample.");

                        report_error(messageService, parent->model(), message);
                    }
                    legacyDistributionFound = true;
                    reader->skipCurrentElement();
                    continue;
                }

                QString tag = reader->attributes().value(Tags::TagAttribute).toString();
                QString displayName =
                    reader->attributes().value(Tags::DisplayNameAttribute).toString();
                if (!topTag.isEmpty()) {
                    // to handle copying of item to another parent
                    tag = topTag;
                    topTag.clear();
                }
                SessionItem* newItem = createItem(parent, model_type, tag);
                if (!newItem) {
                    QString message = QString("Error while parsing XML. Cannot create item of "
                                              "modelType '%1' for tag '%2', parent '%3'")
                                          .arg(model_type, tag, parent->itemName());
                    report_error(messageService, parent->model(), message);
                    // risky attempt to recover the rest of the project
                    reader->skipCurrentElement();
                } else {
                    parent = newItem;
                    parent->setDisplayName(displayName);
                }
                ASSERT(reader->name().toString() == ItemTag);
            } else if (reader->name().toString() == ParameterTag) {
                readProperty(reader, parent);
                gotoEndElementOfTag(reader, ParameterTag);
            } else if (reader->name().toString() == Tags::BinaryData) {
                if (reader->attributes().value(XML::Version).toInt() == 1) {
                    QString valueAsBase64 =
                        reader->readElementText(QXmlStreamReader::SkipChildElements);
                    const auto data = QByteArray::fromBase64(
                        valueAsBase64.toLatin1()); // #baimport add a unit test for this!
                    parent->deserializeBinaryData(data);
                } else
                    throw DeserializationException::tooNew();
                gotoEndElementOfTag(reader, Tags::BinaryData);
            } else if (reader->name().toString() == Tags::NonSessionItemData) {
                parent->readNonSessionItems(reader);
                gotoEndElementOfTag(reader, Tags::NonSessionItemData);
            }
        } else if (reader->isEndElement()) {
            if (reader->name().toString() == XML::ItemTag && parent)
                parent = parent->itemParent();
            if (reader->name() == start_type)
                break;
            if (reader->name() == endTag)
                break;
        }
    }
}

QString XML::readProperty(QXmlStreamReader* reader, SessionItem* item)
{
    ASSERT(reader);
    const QString parameter_name =
        reader->attributes().value(Tags::ParameterNameAttribute).toString();
    const QString parameter_type =
        reader->attributes().value(Tags::ParameterTypeAttribute).toString();
    const int role = reader->attributes().value(Tags::ParameterRoleAttribute).toInt();

    if (!item)
        throw Error(
            QString("Attempt to set property '%1' for non existing item").arg(parameter_name));

    QVariant variant;
    if (parameter_type == "double") {
        double parameter_value =
            reader->attributes().value(Tags::ParameterValueAttribute).toDouble();
        variant = parameter_value;
    } else if (parameter_type == "int") {
        int parameter_value = reader->attributes().value(Tags::ParameterValueAttribute).toInt();
        variant = parameter_value;
    } else if (parameter_type == "uint") {
        unsigned parameter_value =
            reader->attributes().value(Tags::ParameterValueAttribute).toUInt();
        variant = parameter_value;
    } else if (parameter_type == "bool") {
        bool parameter_value = reader->attributes().value(Tags::ParameterValueAttribute).toInt();
        variant = parameter_value;
    } else if (parameter_type == "QString") {
        QString parameter_value =
            reader->attributes().value(Tags::ParameterValueAttribute).toString();
        variant = parameter_value;
    } else if (parameter_type == "QColor") {
        QString parameter_value =
            reader->attributes().value(Tags::ParameterValueAttribute).toString();
        variant = QColor(parameter_value);
    } else if (parameter_type == "ExternalProperty") {
        const bool isMaterialItemColor =
            item->itemParent() != nullptr && dynamic_cast<MaterialItem*>(item->itemParent());

        if (isMaterialItemColor) {
            // This handles legacy XML files for MaterialItem:
            // In former times, the color of a MaterialItem was held as an "ExternalProperty". This
            // has been changed to a simple QColor property => use only the color value of the
            // "ExternalProperty" and store it as QColor in the variant
            const QString colorName =
                reader->attributes().value(Tags::ExternalPropertyColorAtt).toString();
            variant = QColor(colorName);
        } else {
            // This handles legacy XML files for items referring to a material:
            // In former times, the material in e.g. a ParticleItem was held as an
            // "ExternalProperty". This has been changed to the material's identifier only (as
            // QString)
            // => use only the identifier value of the "ExternalProperty" and store it as QString in
            // the variant
            const QString identifier =
                reader->attributes().value(Tags::ExternalPropertyIdentifierAtt).toString();
            variant = identifier;
        }
    } else if (parameter_type == "ComboProperty") {
        QString selections = reader->attributes().value(Tags::ParameterValueAttribute).toString();
        QString values = reader->attributes().value(Tags::ParameterExtAttribute).toString();
        QString tooltips = reader->attributes().value(Tags::ParameterToolAttribute).toString();

        ComboProperty combo_property;
        combo_property.setStringOfValues(values);
        combo_property.setStringOfSelections(selections);
        combo_property.setStringOfTooltips(tooltips);

        variant = combo_property.variant();
    } else {
        QString message = QString("SessionModel::readProperty: parameter type not supported '"
                                  + parameter_type + "'");
        throw Error(message);
    }

    if (variant.isValid()) {
        item->lockModel(true);
        item->setRoleProperty(role, variant);
        item->lockModel(false);
    }

    return parameter_name;
}

SessionItem* XML::readItemAndChildItems(QXmlStreamReader* reader, SessionItem* parent_item,
                                        const QString endTag)
{
    // NOTE: This function is used for the GUI refactoring 'Removal of SessionItem'

    // instantiate a new SessionItem as the root
    SessionItem* newSessionItem = nullptr;
    while (!reader->atEnd()) {
        // first element must be a StartElement
        reader->readNext();

        if (reader->isStartElement()) {
            // interpret the Element based on its name
            // read SessionItem Tag
            if (reader->name().toString() == XML::ItemTag) {
                // extract the model type = the class-name of the new SessionItem
                const QString model_type =
                    reader->attributes().value(Tags::ModelTypeAttribute).toString();

                if (model_type == "ParticleDistribution")
                    throw std::domain_error("ParticleDistribution is not supported");

                // extract the tag = name of the member in the parent class
                QString tag = reader->attributes().value(Tags::TagAttribute).toString();

                newSessionItem = GUI::Model::ItemFactory::CreateItem(model_type);
                if (!newSessionItem) {
                    const QString message =
                        QString("Error while parsing XML. Cannot create item of "
                                "modelType '%1' for tag '%2'")
                            .arg(model_type, tag);
                    throw std::domain_error(message.toStdString());
                }

                QString displayName =
                    reader->attributes().value(Tags::DisplayNameAttribute).toString();
                newSessionItem->setDisplayName(displayName);
                break;
            }
        } else if (reader->isEndElement()) {
            break;
        }
    }


    ASSERT(newSessionItem)

    // inherit the model from the parent
    newSessionItem->setParentAndModel(parent_item, parent_item->model());

    // read the rest of the items until the end tag is reached
    XML::readItems(reader, newSessionItem /* parent */, "" /* topTag */,
                   nullptr /* MessageService */, endTag);
    return newSessionItem;
}


SessionItem* XML::copyItem(const SessionItem* item_to_copy, SessionItem* new_parent)
{
    // NOTE: This function is used for the GUI refactoring 'Removal of SessionItem'

    QByteArray xml_data;
    QXmlStreamWriter writer(&xml_data);
    XML::writeItemAndChildItems(&writer, item_to_copy);

    QXmlStreamReader reader(xml_data);
    // endTag = "" is equivalent to reading the _whole_ XML data
    return XML::readItemAndChildItems(&reader, new_parent, "" /* end tag */);
}
