//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Project/IOService.cpp
//! @brief     Implements class DatafieldIOService
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Project/IOService.h"
#include "GUI/Model/Data/DataItem.h"
#include "GUI/Model/Job/JobItem.h"
#include "GUI/Model/Model/ApplicationModels.h"
#include "GUI/Support/IO/ProjectUtils.h"
#include "GUI/Support/IO/SaveLoadInterface.h"
#include "GUI/Util/MessageService.h"
#include <QFile>
#include <QSet>

namespace {

JobItem* parentJobItem(SaveLoadInterface* item)
{
    auto* cur = dynamic_cast<SessionItem*>(item); // sidecast
    while (cur && cur->modelType() != JobItem::M_TYPE)
        cur = cur->itemParent();

    return dynamic_cast<JobItem*>(cur);
}

} // namespace

DatafieldIOService::DatafieldIOService(QObject* parent)
    : QObject(parent)
    , m_applicationModels(nullptr)
{
    setObjectName("DatafieldIOService");
}

DatafieldIOService::DatafieldIOService(ApplicationModels* models, QObject* parent)
    : QObject(parent)
    , m_applicationModels(nullptr)
{
    setObjectName("DatafieldIOService");
    setApplicationModels(models);
}

void DatafieldIOService::setApplicationModels(ApplicationModels* models)
{
    m_applicationModels = models;
}

void DatafieldIOService::saveDataFiles(const QString& projectDir)
{
    if (!m_history.hasHistory(projectDir))
        m_history.setHistory(projectDir, DatafieldDirHistory());

    DatafieldDirHistory newHistory;

    for (SaveLoadInterface* item : dataInterfaces()) {
        if (m_history.wasModifiedSinceLastSave(projectDir, item))
//            item->saveDatafield(projectDir);
        newHistory.markAsSaved(item);
    }

    // dealing with files
    QStringList oldFiles = m_history.savedFileNames(projectDir);
    QStringList newFiles = newHistory.savedFileNames();
//    cleanOldFiles(projectDir, oldFiles, newFiles);

    // if oldHistory contained some deleted items, that info will be dropped here
    m_history.setHistory(projectDir, newHistory);
}

void DatafieldIOService::loadDataFiles(const QString& projectDir, MessageService* messageService)
{
    DatafieldDirHistory newHistory;
    for (SaveLoadInterface* item : dataInterfaces()) {
        try {
            item->loadDatafield(messageService, projectDir);
            newHistory.markAsSaved(item);
            // handling crash of GUI during job run and non-existing file
            if (auto* jobItem = parentJobItem(item)) {
                if (jobItem->isRunning()) {
                    jobItem->setComments("Possible GUI crash while job was running");
                    jobItem->setStatus(JobStatus::Failed);
                }
            }
        } catch (const std::exception& ex) {
            if (auto* jobItem = parentJobItem(item)) {
                // Handling corrupted file on disk
                jobItem->setComments(
                    QString("Load of the data from disk failed with '%1'").arg(QString(ex.what())));
                jobItem->setStatus(JobStatus::Failed);
            }
            if (!messageService)
                throw ex;
            messageService->addWarning(this, QString(ex.what()));
        }
    }
    m_history.setHistory(projectDir, newHistory);
}

//! Returns all non-XML items available for save/load.

QVector<SaveLoadInterface*> DatafieldIOService::dataInterfaces() const
{
    QVector<SaveLoadInterface*> result;

    if (!m_applicationModels)
        return result;

    for (auto* item : m_applicationModels->dataItems())
        if (auto* data_item = dynamic_cast<SaveLoadInterface*>(item))
            result.push_back(data_item);

    return result;
}

//! Clean old saved files.
//! All files in oldSaves list, which are not in newSaves list, will be removed.

template <typename T>
QSet<T> QListToQSet(const QList<T>& qlist)
{
    return QSet<T>(qlist.constBegin(), qlist.constEnd());
}

void DatafieldIOService::cleanOldFiles(const QString& projectDir, const QStringList& oldSaves,
                                       const QStringList& newSaves)
{
    const auto toRemove = QListToQSet(oldSaves).subtract(QListToQSet(newSaves));
    for (const auto& name : toRemove)
        QFile::remove(projectDir + "/" + name);
}
