//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Project/SaveService.cpp
//! @brief     Implements class SaveService
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Project/SaveService.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Project/ProjectDocument.h"
#include "GUI/Support/IO/ProjectUtils.h"
#include "GUI/Util/Error.h"
#include "GUI/View/Project/AutosaveController.h"
#include <QApplication>
#include <QCoreApplication>
#include <QTime>

namespace {

//! Class to reset flag in the destructor (to ensure resetting even if an exception occurs)
class AutoSetResetFlag {
public:
    AutoSetResetFlag(bool* flag)
        : m_flag(flag)
    {
        *flag = true;
    }
    ~AutoSetResetFlag() { *m_flag = false; }

private:
    bool* m_flag = nullptr;
};

} // namespace

SaveService::SaveService(QObject* parent)
    : QObject(parent)
    , m_is_saving(false)
    , m_autosave(nullptr)
    , m_document(nullptr)
{
}

void SaveService::setDocument(ProjectDocument* document)
{
    m_document = document;

    if (m_autosave)
        m_autosave->setDocument(document);

    m_save_queue.clear();
}

void SaveService::setAutosaveEnabled(bool value)
{
    if (value) {
        delete m_autosave;
        m_autosave = new AutosaveController(this);
        m_autosave->setDocument(m_document);
        connect(m_autosave, &AutosaveController::autosaveRequest, this,
                &SaveService::onAutosaveRequest);
    } else {
        delete m_autosave;
        m_autosave = nullptr;
    }
}

bool SaveService::isAutosaveEnabled() const
{
    return m_autosave;
}

void SaveService::setAutosaveTime(int timerInterval)
{
    if (!m_autosave)
        setAutosaveEnabled(true);

    m_autosave->setAutosaveTime(timerInterval);
}

bool SaveService::isSaving() const
{
    return m_is_saving;
}

void SaveService::stopService()
{
    QApplication::setOverrideCursor(Qt::WaitCursor);

    if (isSaving()) {
        QTime dieTime = QTime::currentTime().addSecs(60);
        while (QTime::currentTime() < dieTime) {
            QCoreApplication::processEvents(QEventLoop::AllEvents, 100);
            if (!isSaving())
                break;
        }
        if (isSaving())
            throw Error("SaveService::stopService() -> Error. Cannot stop service. ");
    }

    if (m_autosave)
        m_autosave->removeAutosaveDir();

    setDocument(nullptr);
    QApplication::restoreOverrideCursor();
}

void SaveService::onAutosaveRequest()
{
    m_document->saveProjectFile(m_autosave->autosaveFullPath(), true);
}
