//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Param/Distrib/ParameterDistribution.cpp
//! @brief     Implements class ParameterDistribution.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Param/Distrib/ParameterDistribution.h"
#include "Base/Util/Assert.h"
#include "Param/Distrib/Distributions.h"
#include "Param/Distrib/ParameterSample.h"

ParameterDistribution::ParameterDistribution(WhichParameter whichParameter,
                                             const IDistribution1D& distribution, size_t n_samples,
                                             double sigma_factor, const RealLimits& limits)
    : m_whichParameter(whichParameter)
    , m_n_samples(n_samples)
    , m_sigma_factor(sigma_factor)
    , m_limits(limits)
    , m_xmin(1.0)
    , m_xmax(-1.0)
{
    m_distribution.reset(distribution.clone());
    if (m_sigma_factor < 0.0)
        throw std::runtime_error("ParameterDistribution::ParameterDistribution() -> Error."
                                 "sigma factor cannot be negative");
    if (n_samples == 0)
        throw std::runtime_error("ParameterDistribution::ParameterDistribution() -> Error."
                                 "Number of samples cannot be zero.");
}

ParameterDistribution::ParameterDistribution(WhichParameter whichParameter,
                                             const IDistribution1D& distribution, size_t n_samples,
                                             double xmin, double xmax)
    : m_whichParameter(whichParameter)
    , m_n_samples(n_samples)
    , m_sigma_factor(0.0)
    , m_xmin(xmin)
    , m_xmax(xmax)
{
    m_distribution.reset(distribution.clone());
    if (m_sigma_factor < 0.0)
        throw std::runtime_error("ParameterDistribution::ParameterDistribution() -> Error."
                                 "sigma factor cannot be negative");
    if (n_samples == 0)
        throw std::runtime_error("ParameterDistribution::ParameterDistribution() -> Error."
                                 "Number of samples cannot be zero.");
    if (xmin >= xmax)
        throw std::runtime_error("ParameterDistribution::ParameterDistribution() -> Error."
                                 "xmin>=xmax");
}

ParameterDistribution::ParameterDistribution(const ParameterDistribution& other)
    : m_whichParameter(other.m_whichParameter)
    , m_n_samples(other.m_n_samples)
    , m_sigma_factor(other.m_sigma_factor)
    , m_limits(other.m_limits)
    , m_xmin(other.m_xmin)
    , m_xmax(other.m_xmax)
{
    m_distribution.reset(other.m_distribution->clone());
}

ParameterDistribution::~ParameterDistribution() = default;

ParameterDistribution& ParameterDistribution::operator=(const ParameterDistribution& other)
{
    if (this != &other) {
        this->m_whichParameter = other.m_whichParameter;
        m_n_samples = other.m_n_samples;
        m_sigma_factor = other.m_sigma_factor;
        m_distribution.reset(other.m_distribution->clone());
        m_limits = other.m_limits;
        m_xmin = other.m_xmin;
        m_xmax = other.m_xmax;
    }
    return *this;
}

std::string ParameterDistribution::whichParameterAsPyEnum() const
{
    switch (whichParameter()) {
    case ParameterDistribution::BeamWavelength:
        return "ParameterDistribution.BeamWavelength";
    case ParameterDistribution::BeamInclinationAngle:
        return "ParameterDistribution.BeamInclinationAngle";
    case ParameterDistribution::BeamAzimuthalAngle:
        return "ParameterDistribution.BeamAzimuthalAngle";
    default:
        throw std::runtime_error("printParameterDistribution(): missing implementation");
    }
}

size_t ParameterDistribution::nDraws() const
{
    if (m_distribution && m_distribution->isDelta())
        return 1;
    return m_n_samples;
}

std::vector<ParameterSample> ParameterDistribution::generateSamples() const
{
    if (m_xmin < m_xmax)
        return m_distribution->equidistantSamplesInRange(m_n_samples, m_xmin, m_xmax);
    return m_distribution->equidistantSamples(m_n_samples, m_sigma_factor, m_limits);
}

std::string ParameterDistribution::unitOfParameter() const
{
    switch (m_whichParameter) {
    case BeamWavelength:
        return "nm";
    case BeamAzimuthalAngle:
    case BeamInclinationAngle:
        return "rad";
    default:
        ASSERT(0);
    }
}
