//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sim/Scan/AlphaScan.cpp
//! @brief     Implements AlphaScan class.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sim/Scan/AlphaScan.h"
#include "Base/Axis/FixedBinAxis.h"
#include "Base/Axis/PointwiseAxis.h"
#include "Device/Beam/IFootprintFactor.h"
#include "Device/Coord/CoordSystem1D.h"
#include "Device/Pol/PolFilter.h"
#include "Param/Distrib/ParameterSample.h"
#include "Param/Distrib/RangedDistributions.h"
#include "Resample/Element/SpecularElement.h"
#include "Sim/Scan/ScanResolution.h"
#include <algorithm>

AlphaScan::AlphaScan(double wl, std::vector<double> inc_angle)
    : ISpecularScan(wl)
    , m_alpha_axis(std::make_unique<PointwiseAxis>("inc_angles", std::move(inc_angle)))
    , m_lambda_distrib(scanEmptyResolution())
    , m_alpha_distrib(scanEmptyResolution())
{
    checkInitialization();
}

AlphaScan::AlphaScan(double wl, const IAxis& inc_angle)
    : ISpecularScan(wl)
    , m_alpha_axis(inc_angle.clone())
    , m_lambda_distrib(scanEmptyResolution())
    , m_alpha_distrib(scanEmptyResolution())
{
    checkInitialization();
}

AlphaScan::AlphaScan(double wl, int nbins, double alpha_i_min, double alpha_i_max)
    : ISpecularScan(wl)
    , m_alpha_axis(std::make_unique<FixedBinAxis>("inc_angles", nbins, alpha_i_min, alpha_i_max))
    , m_lambda_distrib(scanEmptyResolution())
    , m_alpha_distrib(scanEmptyResolution())
{
    checkInitialization();
}

AlphaScan* AlphaScan::clone() const
{
    auto* result = new AlphaScan(m_lambda0, *m_alpha_axis);
    result->setFootprintFactor(m_footprint.get());
    result->setWavelengthResolution(*m_lambda_distrib);
    result->setAngleResolution(*m_alpha_distrib);
    if (m_beamPolarization)
        result->m_beamPolarization.reset(new R3(*m_beamPolarization));
    if (m_polAnalyzer)
        result->m_polAnalyzer.reset(new PolFilter(*m_polAnalyzer));
    return result;
}

AlphaScan::~AlphaScan() = default;

std::vector<SpecularElement> AlphaScan::generateElements() const
{
    std::vector<SpecularElement> result;
    result.reserve(nSteps());

    const DistrOutput lambdaDistrib =
        m_lambda_distrib->generateResolutionSamples(m_lambda0, m_alpha_axis->size());
    const DistrOutput alphaDistrib =
        m_alpha_distrib->generateResolutionSamples(m_alpha_axis->binCenters());

    for (size_t i = 0; i < m_alpha_axis->size(); ++i) {
        for (size_t j = 0; j < alphaDistrib[i].size(); ++j) {
            const double alpha = alphaDistrib[i][j].value;
            for (size_t k = 0; k < lambdaDistrib[i].size(); ++k) {
                const double wavelength = lambdaDistrib[i][k].value;
                const bool computable = wavelength >= 0 && alpha >= 0 && alpha <= M_PI_2;
                result.emplace_back(SpecularElement::FromAlphaScan(
                    wavelength, -alpha, polarizerMatrix(), analyzerMatrix(), computable));
            }
        }
    }
    return result;
}

void AlphaScan::setFootprintFactor(const IFootprintFactor* f_factor)
{
    m_footprint.reset(f_factor ? f_factor->clone() : nullptr);
}

void AlphaScan::setWavelengthResolution(const IScanResolution& resolution)
{
    m_lambda_distrib.reset(resolution.clone());
}

void AlphaScan::setAbsoluteWavelengthResolution(const IRangedDistribution& distr, double std_dev)
{
    std::unique_ptr<IScanResolution> resolution(scanAbsoluteResolution(distr, std_dev));
    setWavelengthResolution(*resolution);
}

void AlphaScan::setAngleResolution(const IScanResolution& resolution)
{
    m_alpha_distrib.reset(resolution.clone());
}

void AlphaScan::setAbsoluteAngularResolution(const IRangedDistribution& distr, double std_dev)
{
    std::unique_ptr<IScanResolution> resolution(scanAbsoluteResolution(distr, std_dev));
    setAngleResolution(*resolution);
}

std::vector<double> AlphaScan::footprint(size_t start, size_t n_elements) const
{
    if (start + n_elements > nSteps())
        throw std::runtime_error("Error in AlphaScan::footprint: given index exceeds the "
                                 "number of simulation elements");

    std::vector<double> result(n_elements, 1.0);
    if (!m_footprint)
        return result;

    const size_t n_wl_samples = m_lambda_distrib->nSamples();
    const size_t n_inc_samples = m_alpha_distrib->nSamples();

    const DistrOutput alphaDistrib =
        m_alpha_distrib->generateResolutionSamples(m_alpha_axis->binCenters());

    const size_t pos_out = start / (n_wl_samples * n_inc_samples);
    size_t pos_inc = (start - pos_out * n_wl_samples * n_inc_samples) / n_wl_samples;
    size_t pos_wl = (start - pos_inc * n_wl_samples);
    int left = static_cast<int>(n_elements);
    size_t pos_res = 0;
    for (size_t i = pos_out; left > 0; ++i) {
        for (size_t k = pos_inc; k < n_inc_samples && left > 0; ++k) {
            pos_inc = 0;
            const double angle = alphaDistrib[i][k].value;
            const double footprint =
                (angle >= 0 && angle <= M_PI_2) ? m_footprint->calculate(angle) : 1.0;
            for (size_t j = pos_wl; j < n_wl_samples && left > 0; ++j) {
                pos_wl = 0;
                result[pos_res] = footprint;
                ++pos_res;
                --left;
            }
        }
    }
    return result;
}

size_t AlphaScan::nSteps() const
{
    return m_alpha_axis->size() * m_lambda_distrib->nSamples() * m_alpha_distrib->nSamples();
}

CoordSystem1D* AlphaScan::scanCoordSystem() const
{
    return new AngularReflectometryCoords(wavelength(), *coordinateAxis());
}

std::vector<double> AlphaScan::createIntensities(const std::vector<SpecularElement>& eles) const
{
    const size_t axis_size = m_alpha_axis->size();
    std::vector<double> result(axis_size, 0.0);

    const DistrOutput lambdaDistrib =
        m_lambda_distrib->generateResolutionSamples(m_lambda0, m_alpha_axis->size());
    const DistrOutput alphaDistrib =
        m_alpha_distrib->generateResolutionSamples(m_alpha_axis->binCenters());

    size_t elem_pos = 0;
    for (size_t i = 0; i < axis_size; ++i) {
        for (size_t j = 0; j < alphaDistrib[i].size(); ++j) {
            const double alpha_wgt = alphaDistrib[i][j].weight;
            for (size_t k = 0; k < lambdaDistrib[i].size(); ++k) {
                const double lambda_wgt = lambdaDistrib[i][k].weight;

                result[i] += eles[elem_pos].intensity() * alpha_wgt * lambda_wgt;
                ++elem_pos;
            }
        }
    }
    return result;
}

void AlphaScan::checkInitialization()
{
    if (m_lambda0 <= 0.0)
        throw std::runtime_error(
            "Error in AlphaScan::checkInitialization: wavelength shell be positive");

    const std::vector<double> axis_values = m_alpha_axis->binCenters();
    if (!std::is_sorted(axis_values.begin(), axis_values.end()))
        throw std::runtime_error("Error in AlphaScan::checkInitialization: q-vector values "
                                 "shall be sorted in ascending order.");

    // TODO: check for inclination angle limits after switching to pointwise resolution.
}
