//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sim/Scan/AlphaScan.h
//! @brief     Declares AlphaScan class.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_SIM_SCAN_ALPHASCAN_H
#define BORNAGAIN_SIM_SCAN_ALPHASCAN_H

#include "Sim/Scan/ISpecularScan.h"
#include <memory>

class ParameterSample;
class IRangedDistribution;
class IScanResolution;

//! Scan type with inclination angles as coordinate values and a unique wavelength.
//! Features footprint correction.
class AlphaScan : public ISpecularScan {
public:
    AlphaScan(double wl, std::vector<double> inc_angle);
    AlphaScan(double wl, const IAxis& inc_angle);
    //! Sets angle-defined specular scan. The first parameter is always a wavelength in nm.
    //! Second parameter is either a numpy array of incident angles in radians or an IAxis object
    //! with angle values. Alternatively an axis can be defined in-place, then
    //! the second passed parameter is the number of bins, third - minimum on-axis angle value,
    //! fourth - maximum on-axis  angle value.
    AlphaScan(double wl, int nbins, double alpha_i_min, double alpha_i_max);
    ~AlphaScan() override;
    AlphaScan* clone() const override;

#ifndef SWIG
    //! Generates simulation elements for specular simulations
    std::vector<SpecularElement> generateElements() const override;

    //! Returns coordinate axis assigned to the data holder
    const IAxis* coordinateAxis() const override
    {
        return m_alpha_axis.get();
    }

    //! Returns IFootprintFactor object pointer
    const IFootprintFactor* footprintFactor() const override
    {
        return m_footprint.get();
    }

    //! Returns footprint correction factor for a range of simulation elements of size _n_elements_
    //! and starting from element with index _i_.
    std::vector<double> footprint(size_t start, size_t n_elements) const override;

    //! Returns the number of simulation elements
    size_t nSteps() const override;

    CoordSystem1D* scanCoordSystem() const override;

    //! Returns intensity vector corresponding to convolution of given simulation elements
    std::vector<double> createIntensities(const std::vector<SpecularElement>& eles) const override;

    // TODO: remove these getters after transition to the new resolution machinery is finished
    const IScanResolution* wavelengthResolution() const
    {
        return m_lambda_distrib.get();
    }
    const IScanResolution* angleResolution() const
    {
        return m_alpha_distrib.get();
    }
#endif // SWIG

    //! Sets footprint correction factor
    void setFootprintFactor(const IFootprintFactor* f_factor);

    //! Sets wavelength resolution values via IScanResolution object.
    void setWavelengthResolution(const IScanResolution& resolution);

    void setAbsoluteWavelengthResolution(const IRangedDistribution& distr, double std_dev);

    //! Sets angle resolution values via IScanResolution object.
    void setAngleResolution(const IScanResolution& resolution);

    void setAbsoluteAngularResolution(const IRangedDistribution& distr, double std_dev);

private:
    using DistrOutput = std::vector<std::vector<ParameterSample>>;

    void checkInitialization();

    const std::unique_ptr<IAxis> m_alpha_axis;
    std::unique_ptr<IFootprintFactor> m_footprint;

    std::unique_ptr<IScanResolution> m_lambda_distrib;
    std::unique_ptr<IScanResolution> m_alpha_distrib;
};

#endif // BORNAGAIN_SIM_SCAN_ALPHASCAN_H
