//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sim/Scan/ScanResolution.cpp
//! @brief     Implements scan resolution class and derived ones.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sim/Scan/ScanResolution.h"
#include "Base/Util/Assert.h"
#include "Param/Distrib/ParameterSample.h"
#include "Param/Distrib/RangedDistributions.h"

namespace {

const std::string relative_resolution = "scanRelativeResolution";
const std::string absolute_resolution = "scanAbsoluteResolution";

//  ************************************************************************************************
//  ScanRelativeResolution
//  ************************************************************************************************

class ScanRelativeResolution : public IScanResolution {
public:
    ScanRelativeResolution(const IRangedDistribution& distr, double reldev)
        : IScanResolution(distr)
        , m_reldev(reldev)
    {
    }
    ~ScanRelativeResolution() override = default;

    IScanResolution* clone() const override
    {
        return new ScanRelativeResolution(*rdistribution(), m_reldev);
    }

    DistrOutput generateResolutionSamples(const std::vector<double>& mean) const override;
    std::vector<double> stdDevs(double mean, size_t n_times) const override;
    std::vector<double> stdDevs(const std::vector<double>& mean) const override;

    double delta() const override { return m_reldev; }

protected:
    std::string resolutionFactoryName() const override { return relative_resolution; }

private:
    double m_reldev; //!< deltas for computing resolutions
};

IScanResolution::DistrOutput
ScanRelativeResolution::generateResolutionSamples(const std::vector<double>& mean) const
{
    ASSERT(!mean.empty());
    DistrOutput result;
    result.reserve(mean.size());
    for (size_t i = 0, size = mean.size(); i < size; ++i)
        result.push_back(rdistribution()->generateSamples(mean[i], mean[i] * m_reldev));
    return result;
}

std::vector<double> ScanRelativeResolution::stdDevs(double mean, size_t n_times) const
{
    return std::vector<double>(n_times, mean * m_reldev);
}

std::vector<double> ScanRelativeResolution::stdDevs(const std::vector<double>& mean) const
{
    ASSERT(!mean.empty());
    std::vector<double> result;
    result.reserve(mean.size());
    for (size_t i = 0, size = mean.size(); i < size; ++i)
        result.push_back(mean[i] * m_reldev);
    return result;
}

//  ************************************************************************************************
//  ScanAbsoluteResolution
//  ************************************************************************************************

class ScanAbsoluteResolution : public IScanResolution {
public:
    ScanAbsoluteResolution(const IRangedDistribution& distr, double stddev)
        : IScanResolution(distr)
        , m_stddev(stddev)
    {
    }
    ~ScanAbsoluteResolution() override = default;

    IScanResolution* clone() const override
    {
        return new ScanAbsoluteResolution(*rdistribution(), m_stddev);
    }

    DistrOutput generateResolutionSamples(const std::vector<double>& mean) const override;
    std::vector<double> stdDevs(double mean, size_t n_times) const override;
    std::vector<double> stdDevs(const std::vector<double>& mean) const override;

    double delta() const override { return m_stddev; }

protected:
    std::string resolutionFactoryName() const override { return absolute_resolution; }

private:
    double m_stddev; //!< deltas for computing resolutions
};

IScanResolution::DistrOutput
ScanAbsoluteResolution::generateResolutionSamples(const std::vector<double>& mean) const
{
    ASSERT(!mean.empty());
    DistrOutput result;
    result.reserve(mean.size());
    for (size_t i = 0, size = mean.size(); i < size; ++i)
        result.push_back(rdistribution()->generateSamples(mean[i], m_stddev));
    return result;
}

std::vector<double> ScanAbsoluteResolution::stdDevs(double, size_t n_times) const
{
    return std::vector<double>(n_times, m_stddev);
}

std::vector<double> ScanAbsoluteResolution::stdDevs(const std::vector<double>& mean) const
{
    ASSERT(!mean.empty());
    return std::vector<double>(mean.size(), m_stddev);
}

//  ************************************************************************************************
//  ScanVectorResolution
//  ************************************************************************************************

class ScanVectorResolution : public IScanResolution {
public:
    ScanVectorResolution(const IRangedDistribution& distr, std::vector<double> stddev)
        : IScanResolution(distr)
        , m_stddev(std::move(stddev))
    {
        ASSERT(!m_stddev.empty());
    }
    ~ScanVectorResolution() override = default;

    IScanResolution* clone() const override
    {
        return new ScanVectorResolution(*rdistribution(), m_stddev);
    }

    DistrOutput generateResolutionSamples(const std::vector<double>& mean) const override;
    std::vector<double> stdDevs(double mean, size_t n_times) const override;
    std::vector<double> stdDevs(const std::vector<double>& mean) const override;

    double delta() const override { return 0; } // TODO

protected:
    std::string resolutionFactoryName() const override { return absolute_resolution; }

private:
    std::vector<double> m_stddev; //!< deltas for computing resolutions
};

IScanResolution::DistrOutput
ScanVectorResolution::generateResolutionSamples(const std::vector<double>& mean) const
{
    const size_t result_size = mean.size();
    ASSERT(result_size == m_stddev.size());
    return rdistribution()->generateSamples(mean, m_stddev);
}

std::vector<double> ScanVectorResolution::stdDevs(double mean, size_t n_times) const
{
    return stdDevs(std::vector<double>(n_times, mean));
}

std::vector<double> ScanVectorResolution::stdDevs(const std::vector<double>& mean) const
{
    ASSERT(mean.size() == m_stddev.size());
    return m_stddev;
}

//  ************************************************************************************************
//  ScanEmptyResolution
//  ************************************************************************************************

class ScanEmptyResolution : public IScanResolution {
public:
    ScanEmptyResolution() = default;

    ~ScanEmptyResolution() override = default;

    IScanResolution* clone() const override { return new ScanEmptyResolution(); }

    DistrOutput generateResolutionSamples(const std::vector<double>& mean) const override;
    std::vector<double> stdDevs(double mean, size_t n_times) const override;
    std::vector<double> stdDevs(const std::vector<double>& mean) const override;

    double delta() const override { return 0; }

protected:
    std::string resolutionFactoryName() const override;
};

ScanEmptyResolution::DistrOutput
ScanEmptyResolution::generateResolutionSamples(const std::vector<double>& mean) const
{
    DistrOutput result;
    result.reserve(mean.size());
    for (size_t i = 0, size = mean.size(); i < size; ++i)
        result.push_back({ParameterSample(mean[i], 1.0)});
    return result;
}

std::vector<double> ScanEmptyResolution::stdDevs(double, size_t n_times) const
{
    return std::vector<double>(n_times, 0.0);
}

std::vector<double> ScanEmptyResolution::stdDevs(const std::vector<double>& mean) const
{
    return std::vector<double>(mean.size(), 0.0);
}

std::string ScanEmptyResolution::resolutionFactoryName() const
{
    ASSERT(0);
}

} // namespace

//  ************************************************************************************************
//  IScanResolution
//  ************************************************************************************************

IScanResolution::~IScanResolution() = default;

size_t IScanResolution::nSamples() const
{
    if (m_distr)
        return m_distr->nSamples();
    return 1;
}

IScanResolution::IScanResolution() = default;

IScanResolution::IScanResolution(const IRangedDistribution& distr)
    : m_distr(distr.clone())
{
}

IScanResolution::DistrOutput IScanResolution::generateResolutionSamples(double mean,
                                                                        size_t n_times) const
{
    return generateResolutionSamples(std::vector<double>(n_times, mean));
}

//  ************************************************************************************************
//  Factory functions
//  ************************************************************************************************

IScanResolution* scanRelativeResolution(const IRangedDistribution& distr, double stddev)
{
    return new ScanRelativeResolution(distr, stddev);
}

IScanResolution* scanAbsoluteResolution(const IRangedDistribution& distr, double stddev)
{
    return new ScanAbsoluteResolution(distr, stddev);
}

IScanResolution* scanVectorResolution(const IRangedDistribution& distr,
                                      const std::vector<double>& stddevs)
{
    return new ScanVectorResolution(distr, stddevs);
}

IScanResolution* scanEmptyResolution()
{
    return new ScanEmptyResolution();
}
