//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sim/Simulation/DepthprobeSimulation.cpp
//! @brief     Implements class DepthprobeSimulation
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sim/Simulation/DepthprobeSimulation.h"
#include "Base/Axis/Bin.h"
#include "Base/Axis/FixedBinAxis.h"
#include "Base/Axis/Frame.h"
#include "Base/Util/Assert.h"
#include "Device/Beam/Beam.h"
#include "Device/Beam/IFootprintFactor.h"
#include "Device/Coord/CoordSystem2D.h"
#include "Device/Data/Datafield.h"
#include "Device/Histo/SimulationResult.h"
#include "Param/Distrib/DistributionHandler.h"
#include "Param/Distrib/Distributions.h"
#include "Resample/Element/DepthprobeElement.h"
#include "Sim/Computation/DepthprobeComputation.h"

namespace {

const RealLimits alpha_limits = RealLimits::limited(0.0, M_PI_2);
const double zero_phi_i = 0.0;
const double zero_alpha_i = 0.0;

} // namespace


DepthprobeSimulation::DepthprobeSimulation(const MultiLayer& sample)
    : ISimulation(sample)
    , m_beam(Beam::horizontalBeam().clone())
{
}

DepthprobeSimulation::~DepthprobeSimulation() = default;

void DepthprobeSimulation::setBeamParameters(double lambda, int nbins, double alpha_i_min,
                                             double alpha_i_max, const IFootprintFactor* beam_shape)
{
    m_alpha_axis = std::make_unique<FixedBinAxis>("alpha_i", static_cast<size_t>(nbins),
                                                  alpha_i_min, alpha_i_max);

    if (lambda <= 0.0)
        throw std::runtime_error(
            "Error in DepthprobeSimulation::setBeamParameters: wavelength must be positive.");
    if (m_alpha_axis->min() < 0.0)
        throw std::runtime_error(
            "Error in DepthprobeSimulation::setBeamParameters: minimum value on "
            "angle axis is negative.");
    if (m_alpha_axis->min() >= m_alpha_axis->max())
        throw std::runtime_error(
            "Error in DepthprobeSimulation::setBeamParameters: alpha bounds are not sorted.");
    if (m_alpha_axis->size() == 0)
        throw std::runtime_error(
            "Error in DepthprobeSimulation::setBeamParameters: angle axis is empty");


    // beam is initialized with zero-valued angles
    // Zero-valued incident alpha is required for proper
    // taking into account beam resolution effects
    beam() = InBeam(beam().intensity(), lambda, zero_alpha_i, zero_phi_i);

    if (beam_shape)
        m_beam->setFootprintFactor(*beam_shape);
}

void DepthprobeSimulation::setzSpan(size_t n_bins, double z_min, double z_max)
{
    if (z_max <= z_min)
        throw std::runtime_error("Error in DepthprobeSimulation::setZSpan: maximum on-axis value "
                                 "is less or equal to the minimum one");
    m_z_axis = std::make_unique<FixedBinAxis>("z", n_bins, z_min, z_max);
}

const ICoordSystem* DepthprobeSimulation::simCoordSystem() const
{
    std::vector<const IAxis*> axes({m_alpha_axis->clone(), m_z_axis->clone()});
    return new DepthprobeCoords(axes, m_beam->ki());
}

std::vector<const INode*> DepthprobeSimulation::nodeChildren() const
{
    std::vector<const INode*> result = ISimulation::nodeChildren();
    result.push_back(m_beam.get());
    return result;
}

//... Overridden executors:

//! init callbacks for setting the parameter values
void DepthprobeSimulation::initDistributionHandler()
{
    for (const auto& distribution : distributionHandler().paramDistributions()) {

        switch (distribution.whichParameter()) {
        case ParameterDistribution::BeamAzimuthalAngle:
            distributionHandler().defineCallbackForDistribution(
                &distribution, [&](double d) { m_beam->setAzimuthalAngle(d); });
            break;
        case ParameterDistribution::BeamInclinationAngle:
            distributionHandler().defineCallbackForDistribution(
                &distribution, [&](double d) { m_beam->setInclination(d); });
            break;
        case ParameterDistribution::BeamWavelength:
            distributionHandler().defineCallbackForDistribution(
                &distribution, [&](double d) { m_beam->setWavelength(d); });
            break;
        default:
            ASSERT(0);
        }
    }
}

void DepthprobeSimulation::initElementVector()
{
    std::vector<DepthprobeElement> result;

    const double wavelength = m_beam->wavelength();
    const double angle_shift = m_beam->alpha_i();

    m_depth_eles.clear();
    m_depth_eles.reserve(alphaAxis()->size());
    for (size_t i = 0; i < alphaAxis()->size(); ++i) {
        double result_angle = m_alpha_axis->bin(i).center() + angle_shift;
        m_depth_eles.emplace_back(wavelength, -result_angle, zAxis(),
                                  alpha_limits.isInRange(result_angle));
    }
}

void DepthprobeSimulation::runComputation(const ReSample& re_sample, size_t iElement)
{
    const auto& begin = m_depth_eles.begin() + static_cast<long>(iElement);
    DepthprobeComputation(re_sample, options(), progress(), begin, begin + static_cast<long>(1))
        .runProtected();
}

void DepthprobeSimulation::normalize(size_t start, size_t n_elements)
{
    const double beam_intensity = m_beam->intensity();
    for (size_t i = start; i < start + n_elements; ++i) {
        DepthprobeElement& element = m_depth_eles[i];
        const double alpha_i = -element.alphaI();
        const auto* footprint = m_beam->footprintFactor();
        double intensity_factor = beam_intensity;
        if (footprint != nullptr)
            intensity_factor = intensity_factor * footprint->calculate(alpha_i);
        element.setIntensities(element.getIntensities() * intensity_factor);
    }
}

void DepthprobeSimulation::addBackgroundIntensity(size_t, size_t)
{
    if (background())
        throw std::runtime_error(
            "Error: nonzero background is not supported by DepthprobeSimulation");
}

void DepthprobeSimulation::addDataToCache(double weight)
{
    const size_t N0 = alphaAxis()->size();
    const size_t N1 = zAxis()->size();
    ASSERT(m_depth_eles.size() == N0);
    ASSERT(m_cache.size() == N0 * N1);

    for (size_t j = 0; j < N0; ++j)
        for (size_t i = 0; i < N1; ++i)
            m_cache[j * N1 + i] += m_depth_eles[j].getIntensities()[i] * weight;
}

void DepthprobeSimulation::validateParametrization(const ParameterDistribution& par_distr) const
{
    const bool zero_mean = par_distr.getDistribution()->mean() == 0.0;
    if (zero_mean)
        return;
    /* TODO come back to this after refactoring of distribution handling
    if (par_distr.whichParameter() == ParameterDistribution::BeamInclinationAngle)
        throw std::runtime_error("Error in DepthprobeSimulation: parameter distribution of "
                                 "beam inclination angle should have zero mean.");
    */
}

//... Overridden getters:

size_t DepthprobeSimulation::nElements() const
{
    return alphaAxis()->size();
}

size_t DepthprobeSimulation::nOutChannels() const
{
    return nElements() * zAxis()->size();
}

SimulationResult DepthprobeSimulation::packResult()
{
    const std::vector<const IAxis*> axes{alphaAxis()->clone(), zAxis()->clone()};
    auto data = std::make_unique<Datafield>(axes, m_cache);

    return {*data, simCoordSystem()};
}

//... Local functions:

const IAxis* DepthprobeSimulation::alphaAxis() const
{
    ASSERT(m_alpha_axis);
    return m_alpha_axis.get();
}

const IAxis* DepthprobeSimulation::zAxis() const
{
    ASSERT(m_z_axis);
    return m_z_axis.get();
}
