//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sim/Simulation/OffspecSimulation.h
//! @brief     Defines class OffspecSimulation.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_SIM_SIMULATION_OFFSPECSIMULATION_H
#define BORNAGAIN_SIM_SIMULATION_OFFSPECSIMULATION_H

#include "Base/Types/OwningVector.h"
#include "Sim/Simulation/ISimulation.h"

class Beam;
class Datafield;
class DiffuseElement;
class IAxis;
class IDetector;
class IPixel;

//! Off-specular scattering simulation.
//!
//! Holds an instrument and sample model.
//! Computes reflected and scattered intensity as function of incident and final glancing angle.
//! @ingroup simulation

class OffspecSimulation : public ISimulation {
public:
    OffspecSimulation(const Beam& beam, const MultiLayer& sample, const IDetector& detector);
    OffspecSimulation(const MultiLayer& sample);
    ~OffspecSimulation() override;

    std::string className() const final { return "OffspecSimulation"; }

    //! Sets beam parameters from here
    void setBeamParameters(double wavelength, const IAxis& alpha_axis, double phi_i);

    IDetector& detector() { return *m_detector; }
    const Beam& beam() const { return *m_beam; }
#ifndef SWIG
    const ICoordSystem* simCoordSystem() const override;

    std::vector<const INode*> nodeChildren() const override;

    //! Returns axis of the beam.
    const IAxis* beamAxis() const;

    Beam& beam()
    {
        return *m_beam;
    }
    const IDetector& detector() const
    {
        return *m_detector;
    }
    const IDetector* getDetector() const
    {
        return m_detector.get();
    }

private:
    //... Overridden executors:
    void initDistributionHandler() override;

    void prepareSimulation() override;

    void initElementVector() override;

    void runComputation(const ReSample& re_sample, size_t iElement) override;

    void normalize(size_t start, size_t n_elements) override;

    void addBackgroundIntensity(size_t start, size_t n_elements) override;

    void addDataToCache(double weight) override;

    //... Overridden getters:
    bool force_polarized() const override;

    size_t nElements() const override;

    SimulationResult packResult() override;

    //! Checks the distribution validity for simulation.
    void validateParametrization(const ParameterDistribution& par_distr) const override;

    ///... Local functions:
    std::vector<DiffuseElement> generateElements(const Beam& beam);

    //... Model components:
    std::shared_ptr<Beam> m_beam;
    std::unique_ptr<IDetector> m_detector;

    //... Caches:
    std::vector<double> m_cache;
    std::vector<DiffuseElement> m_eles;
    std::vector<size_t> m_active_indices; //!< The sequence of bin indices (unmasked, in ROI)
    OwningVector<const IPixel> m_pixels;  //!< All unmasked pixels inside ROI.
    std::unique_ptr<IAxis> m_alpha_i_axis;
#endif // SWIG
};

#endif // BORNAGAIN_SIM_SIMULATION_OFFSPECSIMULATION_H
