//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sim/Simulation/ScatteringSimulation.h
//! @brief     Defines class ScatteringSimulation.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_SIM_SIMULATION_SCATTERINGSIMULATION_H
#define BORNAGAIN_SIM_SIMULATION_SCATTERINGSIMULATION_H

#include "Base/Types/OwningVector.h"
#include "Sim/Simulation/ISimulation.h"

class Beam;
class DiffuseElement;
class IDetector;
class IPixel;
class IShape2D;

//! GISAS simulation.
//!
//! Holds an instrument and sample model.
//! Computes the scattered intensity as function of 2D detector coordinates.
//! @ingroup simulation

class ScatteringSimulation : public ISimulation {
public:
    ScatteringSimulation(const Beam& beam, const MultiLayer& sample, const IDetector& detector);
    ~ScatteringSimulation() override;

    std::string className() const final { return "ScatteringSimulation"; }

    //! Adds mask of given shape to the stack of detector masks. The mask value 'true' means
    //! that the channel will be excluded from the simulation. The mask which is added last
    //! has priority.
    //! @param shape The shape of mask (Rectangle, Polygon, Line, Ellipse)
    //! @param mask_value The value of mask
    void addMask(const IShape2D& shape, bool mask_value = true);

    //! Put the mask for all detector channels (i.e. exclude whole detector from the analysis)
    void maskAll();

    //! Sets rectangular region of interest with lower left and upper right corners defined.
    void setRegionOfInterest(double xlow, double ylow, double xup, double yup);

    Beam& beam() { return *m_beam; }
    IDetector& detector() { return *m_detector; }

#ifndef SWIG
    const ICoordSystem* simCoordSystem() const override;

    const Beam& beam() const
    {
        return *m_beam;
    }
    const IDetector& detector() const
    {
        return *m_detector;
    }
    const IDetector* getDetector() const
    {
        return m_detector.get();
    }

private:
    //... Overridden executors:
    void initDistributionHandler() override;

    void prepareSimulation() override;

    void initElementVector() override;

    void runComputation(const ReSample& re_sample, size_t iElement) override;

    void normalize(size_t start, size_t n_elements) override;

    void addBackgroundIntensity(size_t start, size_t n_elements) override;

    void addDataToCache(double weight) override;

    //... Overridden getters:
    bool force_polarized() const override;

    //! Returns the number of elements this simulation needs to calculate
    size_t nElements() const override;

    SimulationResult packResult() override;

    //... Local function:
    std::vector<DiffuseElement> generateElements(const Beam& beam) const;

    //... Model components:
    std::shared_ptr<Beam> m_beam;
    std::unique_ptr<IDetector> m_detector;

    //... Caches:
    std::vector<DiffuseElement> m_eles;
    std::vector<size_t> m_active_indices; //!< The sequence of bin indices (unmasked, in ROI)
    OwningVector<const IPixel> m_pixels;  //!< All unmasked pixels inside ROI.

#endif // SWIG
};

#endif // BORNAGAIN_SIM_SIMULATION_SCATTERINGSIMULATION_H
