//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sim/Simulation/SpecularSimulation.cpp
//! @brief     Implements class OffspecSimulation.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sim/Simulation/SpecularSimulation.h"
#include "Base/Axis/IAxis.h"
#include "Base/Util/Assert.h"
#include "Device/Beam/IFootprintFactor.h"
#include "Device/Coord/CoordSystem1D.h"
#include "Device/Data/Datafield.h"
#include "Device/Histo/SimulationResult.h"
#include "Param/Distrib/Distributions.h"
#include "Param/Distrib/ParameterSample.h"
#include "Resample/Element/SpecularElement.h"
#include "Resample/Processed/ReSample.h"
#include "Sim/Background/IBackground.h"
#include "Sim/Computation/SpecularComputation.h"
#include "Sim/Scan/AlphaScan.h"

SpecularSimulation::SpecularSimulation(const ISpecularScan& scan, const MultiLayer& sample)
    : ISimulation(sample)
    , m_scan(scan.clone())
{
    // TODO: move inside AlphaScan when pointwise resolution is implemented
    if (scan.coordinateAxis()->min() < 0.0)
        throw std::runtime_error("Invalid scan: minimum value on coordinate axis is negative");
}

SpecularSimulation::~SpecularSimulation() = default;

const ICoordSystem* SpecularSimulation::simCoordSystem() const
{
    return m_scan->scanCoordSystem();
}

//... Overridden executors:

void SpecularSimulation::initElementVector()
{
    m_eles = m_scan->generateElements();
}

void SpecularSimulation::runComputation(const ReSample& re_sample, size_t iElement)
{
    const auto& begin = m_eles.begin() + static_cast<long>(iElement);
    SpecularComputation(re_sample, options(), progress(), begin, begin + static_cast<long>(1))
        .runProtected();
}

void SpecularSimulation::normalize(size_t start, size_t n_elements)
{
    std::vector<double> footprints = m_scan->footprint(start, n_elements);

    for (size_t i = start; i < start + n_elements; ++i) {
        SpecularElement& element = m_eles[i];
        element.setIntensity(element.intensity() * footprints[i - start]);
    }
}

void SpecularSimulation::addBackgroundIntensity(size_t start, size_t n_elements)
{
    if (!background())
        return;
    for (size_t i = start, stop_point = start + n_elements; i < stop_point; ++i) {
        SpecularElement& element = m_eles[i];
        element.setIntensity(background()->addBackground(element.intensity()));
    }
}

void SpecularSimulation::addDataToCache(double weight)
{
    ASSERT(m_eles.size() == m_cache.size());
    for (size_t i = 0; i < m_eles.size(); i++)
        m_cache[i] += m_eles[i].intensity() * weight;
}

//... Overridden getters:

bool SpecularSimulation::force_polarized() const
{
    return m_scan->polarized();
}

size_t SpecularSimulation::nElements() const
{
    return m_scan->nSteps();
}

SimulationResult SpecularSimulation::packResult()
{
    ASSERT(m_eles.size() == m_cache.size());
    for (size_t i = 0; i < m_eles.size(); i++)
        m_eles[i].setIntensity(m_cache[i]);

    ASSERT(!m_eles.empty());
    Datafield data({m_scan->coordinateAxis()->clone()}, m_scan->createIntensities(m_eles));
    return {data, simCoordSystem()};
}

void SpecularSimulation::validateParametrization(const ParameterDistribution& par_distr) const
{
    const bool zero_mean = par_distr.getDistribution()->mean() == 0.0;
    if (zero_mean)
        return;
    /* TODO come back to this after refactoring of distribution handling
    if (par_distr.whichParameter() == ParameterDistribution::BeamInclinationAngle)
        throw std::runtime_error("Error in SpecularSimulation: parameter distribution of "
                                 "beam inclination angle should have zero mean.");
    */
}
