//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Tests/SimFactory/MakeSimulations.cpp
//! @brief     Implements simulation builders for standard tests.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Tests/SimFactory/MakeSimulations.h"
#include "Base/Axis/FixedBinAxis.h"
#include "Base/Const/Units.h"
#include "Device/Beam/Beam.h"
#include "Device/Beam/FootprintGauss.h"
#include "Device/Beam/FootprintSquare.h"
#include "Device/Detector/RectangularDetector.h"
#include "Device/Detector/SphericalDetector.h"
#include "Device/Mask/Ellipse.h"
#include "Device/Mask/Line.h"
#include "Device/Mask/Polygon.h"
#include "Device/Mask/Rectangle.h"
#include "Device/Resolution/ResolutionFunction2DGaussian.h"
#include "Param/Distrib/Distributions.h"
#include "Param/Distrib/RangedDistributions.h"
#include "Resample/Options/SimulationOptions.h"
#include "Sim/Background/ConstantBackground.h"
#include "Sim/Scan/AlphaScan.h"
#include "Sim/Scan/QzScan.h"
#include "Sim/Scan/ScanResolution.h"
#include "Sim/Simulation/includeSimulations.h"
#include <algorithm>
#include <map>
#include <memory>

using Units::angstrom;
using Units::deg;

namespace {

const size_t rdet_nbinsx(40), rdet_nbinsy(30);
const double rdet_width(20.0), rdet_height(18.0), rdet_distance(1000.0);

const R3 zplus(0.0, 0.0, 1.0);
const R3 yplus(0.0, 1.0, 0.0);

const std::map<std::string, std::pair<R3, R3>> ZPolarizationCases = {
    {"PP", {+zplus, +zplus}},
    {"PM", {+zplus, -zplus}},
    {"MP", {-zplus, +zplus}},
    {"MM", {-zplus, -zplus}},
};

const std::map<std::string, std::pair<R3, R3>> YPolarizationCases = {
    {"PP", {+yplus, +yplus}},
    {"PM", {+yplus, -yplus}},
    {"MP", {-yplus, +yplus}},
    {"MM", {-yplus, -yplus}},
};

Beam stdBeam(UnitBeam(1 * angstrom, 0.2 * deg));
Beam powerBeam(InBeam(1e7, 1 * angstrom, 0.2 * deg));

} // namespace

//  ************************************************************************************************
//  GISAS simulations
//  ************************************************************************************************

//! GISAS simulation with small detector and phi[-2,2], theta[0,2].

std::unique_ptr<ScatteringSimulation> test::makeSimulation::MiniGISAS(const MultiLayer& sample)
{
    SphericalDetector detector(25, -2 * deg, 2 * deg, 25, 0, 2 * deg);
    return std::make_unique<ScatteringSimulation>(stdBeam, sample, detector);
}


//! GISAS simulation with small detector and phi[-1,1], theta[0,1].

std::unique_ptr<ScatteringSimulation> test::makeSimulation::MiniGISAS_v2(const MultiLayer& sample)
{
    SphericalDetector detector(25, -1 * deg, 1 * deg, 25, 0, 1 * deg);
    return std::make_unique<ScatteringSimulation>(stdBeam, sample, detector);
}

std::unique_ptr<ScatteringSimulation>
test::makeSimulation::MiniZPolarizedGISAS(const MultiLayer& sample, const std::string& polCase)
{
    const auto zCase = ZPolarizationCases.at(polCase);
    Beam beam(stdBeam);
    beam.setPolarization(zCase.first);
    SphericalDetector detector(25, -2 * deg, 2 * deg, 25, 0, 2 * deg);
    detector.setAnalyzer(zCase.second, 1., .5);
    auto sim = std::make_unique<ScatteringSimulation>(beam, sample, detector);
    return sim;
}


//! Basic GISAS simulation with the detector phi[0,2], theta[0,2].

std::unique_ptr<ScatteringSimulation> test::makeSimulation::BasicGISAS(const MultiLayer& sample)
{
    SphericalDetector detector(100, 0, 2 * deg, 100, 0, 2 * deg);
    return std::make_unique<ScatteringSimulation>(stdBeam, sample, detector);
}

//! Basic GISAS simulation for spin flip channel.

std::unique_ptr<ScatteringSimulation> test::makeSimulation::SpinflipGISAS(const MultiLayer& sample)
{
    Beam beam(stdBeam);
    beam.setPolarization(zplus);
    SphericalDetector detector(100, 0, 2 * deg, 100, 0, 2 * deg);
    detector.setAnalyzer(zplus, -1.0, 0.5);
    return std::make_unique<ScatteringSimulation>(beam, sample, detector);
}

//! GISAS simulation with beam divergence applied.

std::unique_ptr<ScatteringSimulation>
test::makeSimulation::MiniGISASBeamDivergence(const MultiLayer& sample)
{
    std::unique_ptr<ScatteringSimulation> result = MiniGISAS(sample);

    DistributionLogNormal wavelength_distr(1 * angstrom, 0.1);
    DistributionGaussian alpha_distr(0.2 * deg, 0.02 * deg);
    DistributionGate phi_distr(-0.1 * deg, 0.02 * deg);

    result->addParameterDistribution(ParameterDistribution::BeamWavelength, wavelength_distr, 5);
    result->addParameterDistribution(ParameterDistribution::BeamInclinationAngle, alpha_distr, 4);
    result->addParameterDistribution(ParameterDistribution::BeamAzimuthalAngle, phi_distr, 3);

    return result;
}

//! GISAS simulation with multiple masks on the detector plane.

std::unique_ptr<ScatteringSimulation> test::makeSimulation::GISASWithMasks(const MultiLayer& sample)
{
    SphericalDetector detector(50, -1 * deg, 1 * deg, 50, 0, 2 * deg);
    std::unique_ptr<ScatteringSimulation> result =
        std::make_unique<ScatteringSimulation>(powerBeam, sample, detector);

    result->maskAll();
    // pacman
    result->addMask(Ellipse(0.0 * deg, 1.0 * deg, 0.5 * deg, 0.5 * deg), false);
    result->addMask(Ellipse(0.11 * deg, 1.25 * deg, 0.05 * deg, 0.05 * deg), true);

    std::vector<std::pair<double, double>> points = {{0.0 * deg, 1.0 * deg},
                                                     {0.5 * deg, 1.2 * deg},
                                                     {0.5 * deg, 0.8 * deg},
                                                     {0.0 * deg, 1.0 * deg}};
    result->addMask(Polygon(points), true);

    result->addMask(Rectangle(0.45 * deg, 0.95 * deg, 0.55 * deg, 1.05 * deg), false);
    result->addMask(Rectangle(0.61 * deg, 0.95 * deg, 0.71 * deg, 1.05 * deg), false);
    result->addMask(Rectangle(0.75 * deg, 0.95 * deg, 0.85 * deg, 1.05 * deg), false);

    // more masks
    result->addMask(Ellipse(-0.5 * deg, 1.5 * deg, 0.3 * deg, 0.1 * deg, 45. * deg), false);
    result->addMask(VerticalLine(-0.6 * deg), true);
    result->addMask(HorizontalLine(0.3 * deg), false);

    return result;
}

//! GISAS simulation with detector resolution.

std::unique_ptr<ScatteringSimulation>
test::makeSimulation::MiniGISASDetectorResolution(const MultiLayer& sample)
{
    std::unique_ptr<ScatteringSimulation> result = MiniGISAS(sample);
    ResolutionFunction2DGaussian resfunc(0.0025, 0.0025);
    result->detector().setResolutionFunction(resfunc);
    return result;
}

//! GISAS simulation with small detector and including specular peak.

std::unique_ptr<ScatteringSimulation>
test::makeSimulation::MiniGISASSpecularPeak(const MultiLayer& sample)
{
    SphericalDetector detector(25, -2 * deg, 2 * deg, 25, 0, 2 * deg);
    std::unique_ptr<ScatteringSimulation> result =
        std::make_unique<ScatteringSimulation>(stdBeam, sample, detector);
    result->options().setIncludeSpecular(true);
    return result;
}

//! GISAS simulation with large detector to test performance.

std::unique_ptr<ScatteringSimulation> test::makeSimulation::MaxiGISAS(const MultiLayer& sample)
{
    SphericalDetector detector(256, -2 * deg, 2 * deg, 256, 0, 2 * deg);
    return std::make_unique<ScatteringSimulation>(stdBeam, sample, detector);
}

//! Basic GISAS for polarization studies.

std::unique_ptr<ScatteringSimulation> test::makeSimulation::MaxiGISAS00(const MultiLayer& sample)
{
    Beam beam(stdBeam);
    beam.setPolarization(zplus);
    SphericalDetector detector(256, -2 * deg, 2 * deg, 256, 0, 2 * deg);
    detector.setAnalyzer(zplus, 1.0, 0.5);
    return std::make_unique<ScatteringSimulation>(beam, sample, detector);
}

//! GISAS simulation with generic rectangular detector.

std::unique_ptr<ScatteringSimulation>
test::makeSimulation::RectDetectorGeneric(const MultiLayer& sample)
{
    RectangularDetector detector(rdet_nbinsx, rdet_width, rdet_nbinsy, rdet_height);
    detector.setDetectorPosition(R3(rdet_distance, 10.0, 5.0), rdet_width / 2., 1.0,
                                 R3(0.1, -1.0, 0.2));
    return std::make_unique<ScatteringSimulation>(stdBeam, sample, detector);
}

//! GISAS simulation with the rectangular detector perpendicular to the sample.

std::unique_ptr<ScatteringSimulation>
test::makeSimulation::RectDetectorPerpToSample(const MultiLayer& sample)
{
    RectangularDetector detector(rdet_nbinsx, rdet_width, rdet_nbinsy, rdet_height);
    detector.setPerpendicularToSampleX(rdet_distance, rdet_width / 2., 1.0);
    return std::make_unique<ScatteringSimulation>(stdBeam, sample, detector);
}

//! GISAS simulation with the rectangular detector perpendicular to the direct beam.

std::unique_ptr<ScatteringSimulation>
test::makeSimulation::RectDetectorPerpToDirectBeam(const MultiLayer& sample)
{
    RectangularDetector detector(rdet_nbinsx, rdet_width, rdet_nbinsy, rdet_height);
    detector.setPerpendicularToDirectBeam(rdet_distance, rdet_width / 2., 1.0);
    return std::make_unique<ScatteringSimulation>(stdBeam, sample, detector);
}

//! GISAS simulation with the rectangular detector perpendicular to the reflected beam.

std::unique_ptr<ScatteringSimulation>
test::makeSimulation::RectDetectorPerpToReflectedBeam(const MultiLayer& sample)
{
    RectangularDetector detector(rdet_nbinsx, rdet_width, rdet_nbinsy, rdet_height);
    detector.setPerpendicularToReflectedBeam(rdet_distance, rdet_width / 2., 1.0);
    return std::make_unique<ScatteringSimulation>(stdBeam, sample, detector);
}

//! GISAS simulation with Monte-Carlo integration switched ON.

std::unique_ptr<ScatteringSimulation>
test::makeSimulation::MiniGISASMonteCarlo(const MultiLayer& sample)
{
    std::unique_ptr<ScatteringSimulation> result = MiniGISAS(sample);
    result->options().setMonteCarloIntegration(true, 100);
    return result;
}

//! GISAS simulation with spherical detector, region of interest and mask.

std::unique_ptr<ScatteringSimulation>
test::makeSimulation::SphericalDetWithRoi(const MultiLayer& sample)
{
    SphericalDetector detector(40, -2 * deg, 2 * deg, 30, 0, 3 * deg);
    std::unique_ptr<ScatteringSimulation> result =
        std::make_unique<ScatteringSimulation>(stdBeam, sample, detector);
    result->addMask(Rectangle(-0.5 * deg, 0.3 * deg, -0.2 * deg, 0.6 * deg));
    result->setRegionOfInterest(-1.5 * deg, 0.25 * deg, 1.5 * deg, 1.75 * deg);
    return result;
}

//! GISAS simulation with rectangular detector, region of interest and mask.

std::unique_ptr<ScatteringSimulation> test::makeSimulation::RectDetWithRoi(const MultiLayer& sample)
{
    std::unique_ptr<ScatteringSimulation> result = RectDetectorPerpToDirectBeam(sample);
    result->addMask(Rectangle(3.0, 4.0, 5.0, 7.0));
    result->setRegionOfInterest(2.0, 3.0, 18.0, 15.0);
    return result;
}

//! GISAS simulation with an extra long wavelength

std::unique_ptr<ScatteringSimulation>
test::makeSimulation::ExtraLongWavelengthGISAS(const MultiLayer& sample)
{
    Beam beam(InBeam(1e8, 13.52 * Units::nm, 0.2 * deg));
    SphericalDetector detector(100, -1 * deg, 1 * deg, 100, 0.0, 2 * deg);
    auto simulation = std::make_unique<ScatteringSimulation>(beam, sample, detector);
    simulation->options().setIncludeSpecular(true);
    return simulation;
}

//! ISimulation with fitting.
//! Beam intensity set to provide reasonably large values in detector channels.
std::unique_ptr<ScatteringSimulation> test::makeSimulation::MiniGISASFit(const MultiLayer& sample)
{
    SphericalDetector detector(25, -2 * deg, 2 * deg, 25, 0, 2 * deg);
    return std::make_unique<ScatteringSimulation>(powerBeam, sample, detector);
}

//  ************************************************************************************************
//  off-specular simulations
//  ************************************************************************************************

std::unique_ptr<OffspecSimulation> test::makeSimulation::MiniOffspec(const MultiLayer& sample)
{
    auto result = std::make_unique<OffspecSimulation>(sample);

    const int n_alpha(19);
    const double alpha_min(0 * deg);
    const double alpha_max(4 * deg);
    const int n_phi(9);
    const double phi_min(-0.1 * deg);
    const double phi_max(0.1 * deg);

    result->detector().setDetectorParameters(n_phi, phi_min, phi_max, n_alpha, alpha_min,
                                             alpha_max);

    const int n_scan_points(n_alpha);
    const double alpha_i_min(alpha_min);
    const double alpha_i_max(alpha_max);

    FixedBinAxis alpha_i_axis("alpha_i", n_scan_points, alpha_i_min, alpha_i_max);
    result->setBeamParameters(5 * angstrom, alpha_i_axis, 0.0);

    result->beam().setIntensity(1e9);
    result->options().setIncludeSpecular(true);

    return result;
}

//  ************************************************************************************************
//  specular simulations
//  ************************************************************************************************

ISpecularScan* test::makeSimulation::BasicSpecularScan(bool vsQ)
{
    const double wavelength = 1.54 * angstrom;
    const int number_of_bins = 2000;
    const double min_angle = 0;
    const double max_angle = 5 * deg;

    if (vsQ) {
        FixedBinAxis angle_axis("axis", number_of_bins, min_angle, max_angle);
        auto angles = angle_axis.binCenters();
        std::vector<double> qs(angle_axis.size(), 0.0);
        for (size_t i = 0, size = qs.size(); i < size; ++i)
            qs[i] = 4.0 * M_PI * std::sin(angles[i]) / wavelength;
        return new QzScan(qs);
    }
    return new AlphaScan(wavelength, FixedBinAxis("axis", number_of_bins, min_angle, max_angle));
}

std::unique_ptr<SpecularSimulation> test::makeSimulation::BasicSpecular(const MultiLayer& sample,
                                                                        bool vsQ)
{
    std::unique_ptr<ISpecularScan> scan(BasicSpecularScan(vsQ));

    auto result = std::make_unique<SpecularSimulation>(*scan, sample);
    result->options().setUseAvgMaterials(true);
    return result;
}

std::unique_ptr<SpecularSimulation>
test::makeSimulation::BasicYPolarizedSpecular(const MultiLayer& sample, const std::string& polCase,
                                              bool vsQ)
{
    const auto yCase = YPolarizationCases.at(polCase);
    std::unique_ptr<ISpecularScan> scan(BasicSpecularScan(vsQ));
    scan->setPolarization(yCase.first);
    scan->setAnalyzer(yCase.second, 1.0, 0.5);
    return std::make_unique<SpecularSimulation>(*scan, sample);
}

std::unique_ptr<SpecularSimulation>
test::makeSimulation::SpecularWithGaussianBeam(const MultiLayer& sample)
{
    const double wavelength = 1.54 * angstrom;
    const int number_of_bins = 2000;
    const double min_angle = 0;
    const double max_angle = 5 * deg;
    auto gaussian_ff = std::make_unique<FootprintGauss>(1.0);
    AlphaScan scan(wavelength, FixedBinAxis("axis", number_of_bins, min_angle, max_angle));
    scan.setFootprintFactor(gaussian_ff.get());

    return std::make_unique<SpecularSimulation>(scan, sample);
}

std::unique_ptr<SpecularSimulation>
test::makeSimulation::SpecularWithSquareBeam(const MultiLayer& sample)
{
    const double wavelength = 1.54 * angstrom;
    const int number_of_bins = 2000;
    const double min_angle = 0;
    const double max_angle = 5 * deg;
    auto square_ff = std::make_unique<FootprintSquare>(1.0);
    AlphaScan scan(wavelength, FixedBinAxis("axis", number_of_bins, min_angle, max_angle));
    scan.setFootprintFactor(square_ff.get());

    return std::make_unique<SpecularSimulation>(scan, sample);
}

std::unique_ptr<SpecularSimulation>
test::makeSimulation::SpecularDivergentBeam(const MultiLayer& sample)
{
    const double wavelength = 1.54 * angstrom;
    const int number_of_bins = 20;
    const size_t n_integration_points = 10;
    const double min_angle = 0;
    const double max_angle = 5 * deg;
    const double wl_stddev = 0.1 * angstrom;
    const double ang_stddev = 0.1 * deg;
    AlphaScan scan(wavelength, FixedBinAxis("axis", number_of_bins, min_angle, max_angle));

    RangedDistributionGaussian wl_distr(n_integration_points, /*sigma_factor = */ 2.0);
    std::unique_ptr<IScanResolution> wl_res(scanAbsoluteResolution(wl_distr, wl_stddev));

    RangedDistributionGaussian alpha_distr(n_integration_points, /*sigma_factor = */ 2.0);
    std::unique_ptr<IScanResolution> ang_res(scanAbsoluteResolution(alpha_distr, ang_stddev));

    scan.setWavelengthResolution(*wl_res);
    scan.setAngleResolution(*ang_res);

    return std::make_unique<SpecularSimulation>(scan, sample);
}

std::unique_ptr<SpecularSimulation>
test::makeSimulation::TOFRWithRelativeResolution(const MultiLayer& sample)
{
    FixedBinAxis qs("axis", 500, 0.0, 1.0);
    QzScan scan(qs);
    scan.setRelativeQResolution(RangedDistributionGaussian(20, 2.0), 0.03);

    auto result = std::make_unique<SpecularSimulation>(scan, sample);
    result->options().setUseAvgMaterials(true);
    return result;
}

std::unique_ptr<SpecularSimulation>
test::makeSimulation::TOFRWithPointwiseResolution(const MultiLayer& sample)
{
    FixedBinAxis qs("axis", 500, 0.0, 1.0);
    QzScan scan(qs);

    std::vector<double> resolutions;
    resolutions.reserve(qs.size());
    auto qs_vector = qs.binCenters();
    std::for_each(qs_vector.begin(), qs_vector.end(),
                  [&resolutions](double q_val) { resolutions.push_back(0.03 * q_val); });
    scan.setVectorResolution(RangedDistributionGaussian(20, 2.0), resolutions);

    auto result = std::make_unique<SpecularSimulation>(scan, sample);
    result->options().setUseAvgMaterials(true);
    return result;
}

//  ************************************************************************************************
//  depth probe simulations
//  ************************************************************************************************

std::unique_ptr<DepthprobeSimulation>
test::makeSimulation::BasicDepthprobe(const MultiLayer& sample)
{
    const double wavelength = 10 * angstrom;
    const size_t n_alpha = 20;
    const double alpha_min = 0;
    const double alpha_max = 1 * deg;
    const size_t n_z = n_alpha;
    const double z_min = -100.0 * Units::nm;
    const double z_max = 100.0 * Units::nm;

    auto result = std::make_unique<DepthprobeSimulation>(sample);

    result->setBeamParameters(wavelength, n_alpha, alpha_min, alpha_max);
    result->setzSpan(n_z, z_min, z_max);

    return result;
}
