#include "Sim/Scan/ScanResolution.h"
#include "Param/Distrib/ParameterSample.h"
#include "Param/Distrib/RangedDistributions.h"
#include "Tests/GTestWrapper/google_test.h"
#include <cmath>

class IScanResolutionTest : public ::testing::Test {
protected:
    using DistrOutput = std::vector<std::vector<ParameterSample>>;
    void compareResults(const DistrOutput& lhs, const DistrOutput& rhs);
};

void IScanResolutionTest::compareResults(const DistrOutput& lhs, const DistrOutput& rhs)
{
    EXPECT_EQ(lhs.size(), rhs.size());
    for (size_t i = 0; i < lhs.size(); ++i) {
        EXPECT_EQ(lhs[i].size(), rhs[i].size());
        for (size_t j = 0; j < lhs[i].size(); ++j) {
            EXPECT_NEAR(lhs[i][j].value, rhs[i][j].value, std::abs(lhs[i][j].value) * 1e-10);
            EXPECT_NEAR(lhs[i][j].weight, rhs[i][j].weight, std::abs(lhs[i][j].value) * 1e-10);
        }
    }
}

TEST_F(IScanResolutionTest, RelativeSingleValued)
{
    RangedDistributionGate distr(3, 1.0);
    std::unique_ptr<IScanResolution> resolution(scanRelativeResolution(distr, 0.1));

    std::vector<double> coordinates = {1.0, 2.0, 3.0};
    std::vector<double> ref_deltas = {0.1, 0.2, 0.3};
    DistrOutput ref_result = resolution->rdistribution()->generateSamples(coordinates, ref_deltas);
    DistrOutput actual = resolution->generateResolutionSamples(coordinates);
    compareResults(ref_result, actual);

    ref_result = DistrOutput(3, resolution->rdistribution()->generateSamples(1.0, 0.1));
    actual = resolution->generateResolutionSamples(1.0, 3);
    compareResults(ref_result, actual);

    EXPECT_THROW(resolution->generateResolutionSamples(std::vector<double>()), std::runtime_error);
}

TEST_F(IScanResolutionTest, AbsoluteSingleValued)
{
    RangedDistributionGate distr(3, 1.0);
    std::unique_ptr<IScanResolution> resolution(scanAbsoluteResolution(distr, 0.1));

    std::vector<double> coordinates = {1.0, 2.0, 3.0};
    std::vector<double> ref_deltas = {0.1, 0.1, 0.1};
    DistrOutput ref_result = resolution->rdistribution()->generateSamples(coordinates, ref_deltas);
    DistrOutput actual = resolution->generateResolutionSamples(coordinates);
    compareResults(ref_result, actual);

    ref_result = DistrOutput(3, resolution->rdistribution()->generateSamples(1.0, 0.1));
    actual = resolution->generateResolutionSamples(1.0, 3);
    compareResults(ref_result, actual);

    EXPECT_THROW(resolution->generateResolutionSamples(std::vector<double>()), std::runtime_error);
}

TEST_F(IScanResolutionTest, VectorValued)
{
    RangedDistributionGate distr(3, 1.0);
    EXPECT_THROW(scanVectorResolution(distr, std::vector<double>()), std::runtime_error);
    std::unique_ptr<IScanResolution> resolution(scanVectorResolution(distr, {0.2, 0.3, 0.4}));

    std::vector<double> coordinates = {1.0, 2.0, 3.0};
    DistrOutput ref_result =
        resolution->rdistribution()->generateSamples(coordinates, {0.2, 0.3, 0.4});
    DistrOutput actual = resolution->generateResolutionSamples(coordinates);
    compareResults(ref_result, actual);

    ref_result = DistrOutput({resolution->rdistribution()->generateSamples(1.0, 0.2),
                              resolution->rdistribution()->generateSamples(1.0, 0.3),
                              resolution->rdistribution()->generateSamples(1.0, 0.4)});
    actual = resolution->generateResolutionSamples(1.0, 3);
    compareResults(ref_result, actual);

    EXPECT_THROW(resolution->generateResolutionSamples(std::vector<double>()), std::runtime_error);
    EXPECT_THROW(resolution->generateResolutionSamples(std::vector<double>(2, 1.0)),
                 std::runtime_error);
    EXPECT_THROW(resolution->generateResolutionSamples(std::vector<double>(10, 1.0)),
                 std::runtime_error);
}
